# -*- coding: utf-8 -*-
"""


Todo:
    * Use config as much as possible

"""
__author__ = "Anders Åström"
__contact__ = "anders@lyngon.com"
__copyright__ = "2021, Lyngon Pte. Ltd."
__licence__ = """The MIT License
Copyright © 2021 Lyngon Pte. Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the “Software”), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:

The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import sys

from .func import trigger
from .gears import ClosedGearFunction, GearFunction, PartialGearFunction
from .reader import (
    CommandReader,
    GearReader,
    KeysOnlyReader,
    KeysReader,
    PythonReader,
    ShardsIDReader,
    StreamReader,
)
from .sugar import FailurePolicy, KeyType, LogLevel, ReaderType, TriggerMode

__all__ = [
    "trigger",
    "ClosedGearFunction",
    "GearFunction",
    "PartialGearFunction",
    "FailurePolicy",
    "KeyType",
    "LogLevel",
    "ReaderType",
    "TriggerMode",
    "utils",
    "CommandReader",
    "GearReader",
    "KeysOnlyReader",
    "KeysReader",
    "PythonReader",
    "ShardsIDReader",
    "StreamReader",
]

# Dynamic / Conditional imports below. (depends on installed packa)

try:
    # This will fail if redis package is not installed
    from .client import Gears, Redis

    __all__ += ["Gears", "Redis"]

    try:
        # This will fail if redis-py-cluster package is not installed
        from .cluster import RedisCluster, RedisGears

        __all__ += ["RedisCluster"]
    except ModuleNotFoundError:
        from .client import Redis as RedisGears

    __all__ += ["RedisGears"]

except ModuleNotFoundError:
    pass


try:
    # This will fail if redis package is not installed
    from .command import cmd

    __all__ += ["cmd"]
except ModuleNotFoundError:
    pass


try:
    # this will fail if either redis, watchdog or pyaml packages are not installed
    from .loader import GearsLoader

    __all__ += ["GearsLoader"]
except ModuleNotFoundError:
    pass

try:
    # this will fail if packaging packages is not installed
    from .requirements import (
        Requirement,
        Version,
        read_requirements,
        resolve_requirements,
    )

    __all__ += [
        "Requirement",
        "Version",
        "read_requirements",
        "resolve_requirements",
    ]
except ModuleNotFoundError:
    pass

# Use either the real or mock (placeholder) implementations of the
# Redis Gears Python environment top level builtin funvtions
# Depending on if the module is loaded in a 'redisgears' environment
# or not (e.g. dev or client)
if "redisgears" in sys.modules:
    # Server Gears runtime environment
    GEARS_RUNTIME = True
    # Import the default functions and classes
    # pyright: reportMissingImports=false
    from __main__ import GB as GB
    from __main__ import GearsBuilder as GearsBuilder
    from __main__ import configGet as configGet
    from __main__ import gearsConfigGet as gearsConfigGet
    from redisgears import atomicCtx as atomic
    from redisgears import executeCommand as execute
    from redisgears import getMyHashTag as hashtag
    from redisgears import log as log

else:
    # Dev or Client environment
    GEARS_RUNTIME = False
    # Import placeholder functions and
    from .runtime import (
        GB,
        GearsBuilder,
        atomic,
        configGet,
        execute,
        gearsConfigGet,
        hashtag,
        log,
    )

__all__ += [
    "GB",
    "GearsBuilder",
    "atomic",
    "configGet",
    "execute",
    "gearsConfigGet",
    "hashtag",
    "log",
    "GEARS_RUNTIME",
]



# -*- coding: utf-8 -*-
"""
Helper module for the sole purpose of handling log formatting.
"""
__author__ = "Anders Åström"
__contact__ = "anders@lyngon.com"
__copyright__ = "2021, Lyngon Pte. Ltd."
__licence__ = """The MIT License
Copyright © 2021 Lyngon Pte. Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the “Software”), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:

The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import configparser
import json
import logging
import logging.config
import pathlib
from datetime import datetime
from typing import Any, Dict, Union

import yaml
import yaml.error

configparser.MissingSectionHeaderError

iso8601_datefmt = "%Y-%m-%dT%H:%M:%S.%f"
"""ISO 8601 date / time format string."""


# Logging matters
class UTC_ISO8601_Formatter(logging.Formatter):
    """Formatter for UTC timestamps in ISO 8601 date format."""

    converter = datetime.utcfromtimestamp  # type: ignore

    def formatTime(self, record, datefmt=None):
        ct = self.converter(record.created)
        if datefmt:
            s = ct.strftime(datefmt)
        else:
            s = ct.strftime(iso8601_datefmt)
        return s


log_message_format = (
    "%(asctime)s\t"
    "%(levelname)s\t"
    "%(name)s.%(funcName)s@%(lineno)d:\t"
    "%(message)s\t"
    "EOL"
)
"""Default log message format. E.g:

"""

log_formatter = UTC_ISO8601_Formatter(fmt=log_message_format)
"""Default log formatter"""


# TODO: Simplify
def initialize_logger(  # noqa: C901
    level=logging.DEBUG,
    fmt: Union[logging.Formatter, str] = None,
    config: Union[pathlib.Path, str, Dict[str, Any]] = None,
) -> None:
    """Initialize logger with defaults, overridden by any custom options from various
    sources, including passed parameters, json, ini or yaml config file.

    Args:
        fmt (Union[logging.Formatter, str], optional):
            Logging format string or Formatter to use.
            Defaults to the default log message format:

                ``2021-03-08T10:46:15.1329  DEBUG   package.module.func@123
                    This is a message   EOL``

        config (Union[pathlib.Path, str, Dict[str, Any]], optional):
            Optional ini, json or yaml filepath or a dict with config parameters for
            the logger.
            Defaults to None.

    Raises:
        FileNotFoundError:
            Raised if the config option is a path to a file that does not exist.

        ValueError:
            Raised if the config option is an existing file but cannot be parsed.
    """
    # Initialize to good (tm) defaults
    logging.basicConfig(level=level, format=log_message_format, datefmt=iso8601_datefmt)

    if not fmt:
        fmt = log_message_format

    if not isinstance(fmt, logging.Formatter):
        msg_formatter = UTC_ISO8601_Formatter(fmt=str(fmt))

    # Override with explicit opthions
    root_logger = logging.getLogger()
    root_logger.setLevel(level)
    root_logger.handlers[0].setFormatter(msg_formatter)

    # Override with config options, if any
    if not config:
        return

    if isinstance(config, dict):
        logging.config.dictConfig(config)
        logging.getLogger(__name__).info("Successfully loaded configuration from dict")
        return

    conf_file = pathlib.Path(str(config))
    suffix = conf_file.suffix.lower()
    config_dict = None

    if not conf_file.is_file():
        raise FileNotFoundError(conf_file)

    parse_error_string = "Unable to parse config file as {}: '{}'"

    # assume yaml, unelss explicitly otherwise
    if suffix not in [".ini", ".json"]:
        logging.getLogger(__name__).debug(
            f"Attempting to parse '{conf_file}' as yaml file."
        )
        try:
            with conf_file.open() as f:
                config_dict = yaml.safe_load(f.read())
                logging.getLogger(__name__).info(f"Successfully read '{conf_file}''")
        except yaml.error.YAMLError as e:
            msg = parse_error_string.format("YAML", conf_file)
            if suffix in [".yaml", ".yml"]:
                raise ValueError(msg) from e
            else:
                logging.getLogger(__name__).warning(msg)

    # if not yaml, assume json, unless explicitly otherwise
    if not config_dict and suffix not in [".ini", ".yaml", ".yml"]:
        logging.getLogger(__name__).debug(
            f"Attempting to parse '{conf_file}' as json file."
        )
        try:
            with conf_file.open() as f:
                config_dict = json.load(f)
                logging.getLogger(__name__).info(f"Successfully read '{conf_file}'")

        except json.JSONDecodeError as e:
            msg = parse_error_string.format("JSON", conf_file)
            if suffix == ".json":
                raise ValueError(msg) from e
            else:
                logging.getLogger(__name__).warning(msg)

    # if neither json nor yaml, attempt to parse as legacy ini
    if not config_dict and suffix not in [".json", ".yaml", ".yml"]:
        logging.getLogger(__name__).debug(
            f"Attempting to parse '{conf_file}' as legacy ini file."
        )
        try:
            logging.config.fileConfig(str(conf_file), disable_existing_loggers=False)
            logging.getLogger(__name__).info(f"Successfully loaded '{conf_file}'")
            return
        except configparser.Error as e:
            msg = parse_error_string.format("INI", conf_file)
            if suffix == ".ini":
                raise ValueError(msg) from e
            else:
                logging.getLogger(__name__).warning(msg)

    if not config_dict:
        msg = parse_error_string.format("anything", conf_file)
        raise ValueError(msg)

    logging.config.dictConfig(config_dict)
    logging.getLogger(__name__).info("Successfully loaded configuration")
    return



# -*- coding: utf-8 -*-
"""
Command line interface of redgrease.
Allows for loading of gears python scripts into a Redis Gears cluster.
"""
__author__ = "Anders Åström"
__contact__ = "anders@lyngon.com"
__copyright__ = "2021, Lyngon Pte. Ltd."
__licence__ = """The MIT License
Copyright © 2021 Lyngon Pte. Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the “Software”), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:

The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import logging
import pathlib
import re
import time

import configargparse

import redgrease.formatting
import redgrease.loader

args = configargparse.ArgParser(
    description="Scans one or more directories for Redis Gears scripts, and "
    "executes them in a Redis Gears instance or cluster. "
    "Can optionally run continiously, montoring and re-loading scripts "
    "whenever changes are detected.",
    default_config_files=["./*.conf", "/etc/redgrease/conf.d/*.conf"],
)

args.add_argument(
    "directories",
    metavar="dir_path",
    nargs="+",
    type=pathlib.Path,
    help="One or more directories containing Redis Gears scripts to watch",
)

args.add_argument(
    "-c",
    "--config",
    env_var="CONFIG_FILE",
    metavar="PATH",
    is_config_file=True,
    help="Config file path",
)
args.add_argument(
    "--index-prefix",
    env_var="INDEX_PREFIX",
    metavar="PREFIX",
    required=False,
    type=str,
    default=redgrease.loader.default_index_prefix,
    help="Redis key prefix added to the index of monitored/executed script " "files.",
)
args.add_argument(
    "-r",
    "--recursive",
    env_var="RECURSIVE",
    action="store_true",
    help="Recursively watch subdirectories.",
)
args.add_argument(
    "--script-pattern",
    env_var="SCRIPT_PATTERN",
    metavar="PATTERN",
    required=False,
    type=str,
    default=redgrease.loader.default_script_pattern,
    help="File name pattern (glob-style) that must be matched for scripts to "
    "be loaded.",
)
args.add_argument(
    "--requirements-pattern",
    env_var="REQUIREMENTS_PATTERN",
    metavar="PATTERN",
    required=False,
    type=str,
    default=redgrease.loader.default_requirements_pattern,
    help="File name pattern (glob-style) that must be matched for requirement "
    "files to be loaded.",
)
args.add_argument(
    "--unblocking-pattern",
    env_var="UNBLOCKING_PATTERN",
    metavar="PATTERN",
    required=False,
    type=re.compile,
    default=redgrease.loader.default_unblocking_pattern,
    help="Scripts with file paths that match this regular expression, "
    "will be executed with the 'UNBLOCKING' modifier, i.e. async execution. "
    "Note that the pattern is a 'search' pattern and not anchored to the"
    "start of the path string.",
)
args.add_argument(
    "-i",
    "--ignore",
    env_var="IGNORE",
    metavar="PATTERN",
    action="append",
    required=False,
    help="Ignore files matching this pattern.",
)
args.add_argument(
    "-w",
    "--watch",
    env_var="WATCH",
    metavar="SECONDS",
    nargs="?",
    type=float,
    default=0.0,
    const=5.0,
    help="If set, the directories will be continously montiored for updates/"
    "modifications to scripts and requirement files, and automatically loaded/"
    "rerun. The flag takes an optional value specifying the  duration, "
    "in seconds, to wait for further updates/modifications to files, before "
    "executing. "
    "This 'hysteresis' period is to prevent malformed scripts to be "
    "unnecessarily loaded during coding. "
    "If no value is supplied, the duration is defaulting to 5 seconds.",
)
args.add_argument(
    "-s",
    "--server",
    env_var="SERVER",
    type=str,
    const="redis",
    default="localhost",
    nargs="?",
    help="Redis Gears host server IP or hostname.",
)
args.add_argument(
    "-p",
    "--port",
    env_var="PORT",
    type=int,
    default=6379,
    help="Redis Gears host port number",
)
args.add_argument(
    "-l",
    "--log-config",
    env_var="LOG_CONFIG",
    type=str,
    help="",
)


config = args.parse_args()

redgrease.formatting.initialize_logger(config=config.log_config)

log = logging.getLogger(__name__)


def main():
    log.debug(config)
    loader = redgrease.loader.GearsLoader(
        script_pattern=config.script_pattern,
        requirements_pattern=config.requirements_pattern,
        unblocking_pattern=config.unblocking_pattern,
        ignore_patterns=config.ignore,
        index_prefix=config.index_prefix,
        server=config.server,
        port=config.port,
        observe=config.watch,
    )
    for directory in config.directories:
        loader.add_directory(directory=directory, recursive=config.recursive)

    if loader.observer:
        log.info("Starting directory observer!")
        loader.start()
        running = True

        try:
            while running:
                time.sleep(1)
        except KeyboardInterrupt:
            log.warning("Interrupted by user. Ending!")
        finally:
            loader.stop()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Client library module for Redis Gears servers, exposing the gears-specific commands
Can be instantiated in a few ways:

Examples:
    - As a redis client::

        import redgrease

        r = redgrease.RedisGears()  # Takes same arguments as redis.Redis
        r.gears.pyexecute(...)


    - As a separate Gears object, taking a redis.Redis as parameter::

        import redis
        import redgrease

        r = redis.Redis()
        g = redgrease.Gears(r)
        g.pyexecute(...)

"""
__author__ = "Anders Åström"
__contact__ = "anders@lyngon.com"
__copyright__ = "2021, Lyngon Pte. Ltd."
__licence__ = """The MIT License
Copyright © 2021 Lyngon Pte. Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the “Software”), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:

The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import logging
import os.path
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Union

import redis

import redgrease.config
import redgrease.data
import redgrease.gears
import redgrease.reader
import redgrease.requirements
import redgrease.runtime
from redgrease.utils import (
    CaseInsensitiveDict,
    bool_ok,
    dict_of,
    list_parser,
    safe_str,
    to_redis_type,
)

log = logging.getLogger(__name__)

ExecutionID = Union[bytes, str, redgrease.data.ExecID, redgrease.data.ExecutionInfo]
"""Type alias for valid execution identifiers"""

RegistrationID = Union[bytes, str, redgrease.data.ExecID, redgrease.data.Registration]
"""Type alias for valid registration identifiers"""


class Gears:
    """Client class for Redis Gears commands.

    Attributes:
        redis (redis.Redis):
            Redis client / connection, used for the underlying communication with
            the server.

        config (redgrease.config.Config):
            Redis Gears Configuration 'client'
    """

    RESPONSE_CALLBACKS: Dict[str, Callable] = {
        "RG.ABORTEXECUTION": bool_ok,
        "RG.CONFIGGET": dict_of(
            CaseInsensitiveDict(redgrease.config.Config.ValueTypes)
        ),
        "RG.CONFIGSET": lambda res: all(map(bool_ok, res)),
        "RG.DROPEXECUTION": bool_ok,
        "RG.DUMPEXECUTIONS": list_parser(redgrease.data.ExecutionInfo.from_redis),
        "RG.DUMPREGISTRATIONS": list_parser(redgrease.data.Registration.from_redis),
        "RG.GETEXECUTION": redgrease.data.ExecutionPlan.parse,
        "RG.GETRESULTS": redgrease.data.parse_execute_response,
        "RG.GETRESULTSBLOCKING": redgrease.data.parse_execute_response,
        "RG.INFOCLUSTER": redgrease.data.ClusterInfo.parse,
        "RG.PYEXECUTE": redgrease.data.parse_execute_response,
        "RG.PYSTATS": redgrease.data.PyStats.from_redis,
        "RG.PYDUMPREQS": list_parser(redgrease.data.PyRequirementInfo.from_redis),
        "RG.REFRESHCLUSTER": bool_ok,
        "RG.TRIGGER": redgrease.data.parse_trigger_response,
        "RG.UNREGISTER": bool_ok,
    }

    def __init__(self, redis: redis.Redis):
        """Instatiate a Gears client objeect

        Args:
            redis (redis.Redis):
                redis.Redis client object, used for the underlying communication with
                the server.
        """
        self.redis = redis
        self.config = redgrease.config.Config(redis)

    def _ping(self) -> bool:
        """Test server liveness/connectivty

        Returns:
            bool:
                True if, and only if, the Redis server is responsive and is running the
                gears module
        """
        return self.config.MaxExecutions > 0

    def abortexecution(self, id: ExecutionID) -> bool:
        """Abort the execution of a function mid-flight

        Args:
            id (Union[redgrease.data.ExecutionInfo, redgrease.data.ExecID, bytes, str]):
                The execution id to abort

        Returns:
            bool:
                True or an error if the execution does not exist or had already
                finished.
        """
        if isinstance(id, redgrease.data.ExecutionInfo):
            id = id.executionId

        return self.redis.execute_command("RG.ABORTEXECUTION", to_redis_type(id))

    def dropexecution(self, id: ExecutionID) -> bool:
        """Remove the execution of a function from the executions list.

        Args:
            id (Union[redgrease.data.ExecutionInfo, redgrease.data.ExecID, bytes, str]):
                Execution ID to remove

        Returns:
            bool:
                True if successful, or an error if the execution does not exist or is
                still running.
        """
        if isinstance(id, redgrease.data.ExecutionInfo):
            id = id.executionId
        return self.redis.execute_command("RG.DROPEXECUTION", to_redis_type(id))

    def dumpexecutions(self) -> List[redgrease.data.ExecutionInfo]:
        """Get list of function executions.
        The executions list's length is capped by the 'MaxExecutions' configuration
        option.

        Returns:
            List[redgrease.data.ExecutionInfo]:
                A list of ExecutionInfo, with an entry per execution.
        """
        return self.redis.execute_command("RG.DUMPEXECUTIONS")

    def dumpregistrations(self) -> List[redgrease.data.Registration]:
        """Get list of function registrations.

        Returns:
            List[redgrease.data.Registration]:
                A list of Registration, with one entry per registered function.
        """
        return self.redis.execute_command("RG.DUMPREGISTRATIONS")

    def getexecution(
        self,
        id: ExecutionID,
        locality: Optional[redgrease.data.ExecLocality] = None,
    ) -> Mapping[bytes, redgrease.data.ExecutionPlan]:
        """Get the executoin plan details for a function in the execution list.

        Args:
            id (Union[redgrease.data.ExecutionInfo, redgrease.data.ExecID, bytes, str]):
                Execution identifier for the function to fetch execution plan for.

            locality (Optional[redgrease.data.ExecLocality], optional):
                Set to 'Shard' to get only local execution pland and set to 'Cluster'
                to collect executions from all shards.
                Defaults to 'Shard' in stand-alone mode, but "Cluster" in cluster mode.

        Returns:
            Mapping[bytes, redgrease.data.ExecutionPlan]:
                A dict, mapping cluster ID to ExecutionPlan
        """
        if isinstance(id, redgrease.data.ExecutionInfo):
            id = id.executionId

        loc = [] if locality is None else [safe_str(locality).upper()]

        return self.redis.execute_command("RG.GETEXECUTION", to_redis_type(id), *loc)

    def getresults(
        self,
        id: ExecutionID,
    ) -> redgrease.data.ExecutionResult:
        """Get the results of a function in the execution list.

        Args:
            id (Union[redgrease.data.ExecutionInfo, redgrease.data.ExecID, bytes, str]):
                Execution identifier for the fuction to fetch the output for.

        Returns:
            redgrease.data.ExecutionResult:
                Results and errors from the gears function, if, and only if, execution
                exists and is completed.

        Raises:
            redis.exceptions.ResponseError:
                If the the execution does not exist or is still running
        """

        if isinstance(id, redgrease.data.ExecutionInfo):
            id = id.executionId

        return self.redis.execute_command("RG.GETRESULTS", to_redis_type(id))

    def getresultsblocking(self, id: ExecutionID) -> redgrease.data.ExecutionResult:
        """Get the results and errors from the execution details of a function.
        If the execution is not finished, the call is blocked until execution
        ends.

        Args:
            id (Union[redgrease.data.ExecutionInfo, redgrease.data.ExecID, bytes, str]):
                Execution identifier for the fuction to fetch the results and errors
                for.

        Returns:
            redgrease.data.ExecutionResult:
                Results and errors from the gears function if the execution exists.

        Raises:
            redis.exceptions.ResponseError:
                If the the execution does not exist.
        """
        if isinstance(id, redgrease.data.ExecutionInfo):
            id = id.executionId

        return self.redis.execute_command("RG.GETRESULTSBLOCKING", to_redis_type(id))

    def infocluster(self) -> redgrease.data.ClusterInfo:
        """Gets information about the cluster and its shards.

        Returns:
            redgredase.data.ClusterInfo:
                Cluster information or None if not in cluster mode.
        """
        return self.redis.execute_command("RG.INFOCLUSTER")

    def pyexecute(
        self,
        gear_function: Union[
            str, redgrease.runtime.GearsBuilder, redgrease.gears.GearFunction
        ] = "",
        unblocking=False,
        requirements: Optional[
            Iterable[Union[str, redgrease.requirements.Requirement]]
        ] = None,
        enforce_redgrease: redgrease.requirements.PackageOption = None,
    ) -> redgrease.data.ExecutionResult:
        """Execute a gear function.

        Args:
            gear_function (Union[str, redgrease.gears.GearFunction], optional):
                - A string containgg a clear-text serialized Gears Python function as
                    per the official documentation.
                    (https://oss.redislabs.com/redisgears/intro.html#the-simplest-example)

                - A GearsBuilder or GearFunction object. Notes:
                    * Python version must match the Gear runtime.
                    * If the function is not "closed" with a `run()` or `register()`
                    operation, an `run()` operation without arguments will be assumed,
                    and automatically added to the function to close it.
                    * The default for `enforce_redgrease` is True.

                - A file path to a gear script. This script can

                Defaults to "" (no function).

            unblocking (bool, optional):
                Execute function without waiting for it to finish, before returnining.
                Defaults to False. I.e. block until the function returns or fails.

            requirements (Iterable[Union[None, str, redgrease.requirements.Requirement]], optional):
                List of 3rd party package requirements needed to execute the function
                on the server.
                Defaults to None.

            enforce_redgrease (redgrease.requirements.PackageOption, optional):
                Indicates if redgrease runtime package requirement should be added or not,
                and potentially which version and/or extras or source.
                It can take several optional types::

                    - None :  no enforcing. Requirements are passed through,
                            with or without 'redgrease'.

                    - bool :
                        True - enforces latest 'redgrease[runtime] ' package on PyPi,
                        False - enforces that 'redgrease' is NOT in the requirements,
                            any redgrease requirements will be removed from the function.
                            Note that it will not force redgrease to be uninstalled.

                    - str :
                        a. Specific version. E.g. "1.2.3".
                        b. Version qualifier. E.g. ">=1.0.0"
                        c. Extras. E.g. "all" or "runtime".
                            Will enforce the latest version on PyPi, with this/these extras
                        d. Full requirement qualifier or source. E.g:
                            "redgrease[all]>=1.2.3"
                            "redgrease[runtime]@git+https://github.com/lyngon/redgrease.git@main"

                    - Version : behaves just as string version (a.)

                    - Requirement : behaves just as string version (d.)

                Defaults to False (for str function), True for GearFunction objects.

        Returns:
            redgrease.data.ExecutionResult:
                The value of the ExecutionResult will contain the result of the function, unless:

                - When used in 'unblocking' mode, the value is set to the execution ID

                - If the function has no output (i.e. it is closed by `register()`
                    or for some other reason is not closed by a `run()` action),
                    the value is True (boolean).

                Any results and errors generated by the function are accumulated in the results'
                'errors' property, as a list.

        Raises:
            redis.exceptions.ResponseError:
                If the funvction cannot be parsed.
        """  # noqa

        requirements = set(requirements if requirements else [])
        result_function = None

        if isinstance(gear_function, redgrease.runtime.GearsBuilder):
            gear_function = gear_function._function

        if isinstance(gear_function, redgrease.gears.GearFunction):
            # If the input is a GearFunction, we get the requirements from it,
            # and ensure that redgrease is included
            requirements = requirements.union(gear_function.requirements)
            enforce_redgrease = enforce_redgrease or True

            if isinstance(gear_function, redgrease.gears.PartialGearFunction):
                # If the function isn't closed with either 'run' or 'register'
                # we assume it is meant to be closed with a 'run'
                gear_function = gear_function.run()

            function_string = redgrease.data.seralize_gear_function(gear_function)

            # Special case for CommandReader functions:
            # return a new function that calls its "trigger" and returns the results.
            if gear_function.reader == "CommandReader":

                def result_function(*args):
                    return self.trigger(
                        gear_function.operation.kwargs["trigger"], *args
                    )

        elif os.path.exists(gear_function):
            # If the gear function is a fle path,
            # then we load the contents of the file
            with open(gear_function) as script_file:
                function_string = script_file.read()

        else:
            # Otherwise we default to the string version of the function
            function_string = str(gear_function)

        params = []
        if unblocking:
            params.append("UNBLOCKING")

        # Resolve requirement conflicts, remove duplicates
        requirements = redgrease.requirements.resolve_requirements(
            requirements, enforce_redgrease=enforce_redgrease
        )

        if requirements:
            params.append("REQUIREMENTS")
            params += list(map(str, requirements))

        command_response = self.redis.execute_command(
            "RG.PYEXECUTE",
            function_string,
            *params,
        )

        if result_function and command_response:
            return redgrease.data.ExecutionResult(result_function)

        return command_response

    def pystats(self) -> redgrease.data.PyStats:
        """Gets memory usage statisticy from the Python interpreter

        Returns:
            redgrease.data.PyStats:
                Python interpretere memory statistics, including total,
                peak and current amount of allocated memory, in bytes.
        """
        return self.redis.execute_command("RG.PYSTATS")

    def pydumpreqs(self) -> List[redgrease.data.PyRequirementInfo]:
        """Gets all the python requirements available (with information about
        each requirement).

        Returns:
            List[redgrease.data.PyRequirementInfo]:
                List of Python requirement information objects.
        """
        return self.redis.execute_command("RG.PYDUMPREQS")

    def refreshcluster(self) -> bool:
        """Refreshes the local node's view of the cluster topology.

        Returns:
            bool:
                True if successful.

        Raises:
            redis.exceptions.ResponseError:
                If not successful
        """
        return self.redis.execute_command("RG.REFRESHCLUSTER")

    def trigger(self, trigger_name: str, *args) -> List[Any]:
        """Trigger the execution of a registered 'CommandReader' function.

        Args:
            trigger_name (str):
                The registered 'trigger' name of the function

            *args (Any):
                Any additional arguments to the trigger

        Returns:6
            List: A list of the functions output records.
        """
        return self.redis.execute_command("RG.TRIGGER", safe_str(trigger_name), *args)

    def unregister(self, id: RegistrationID) -> bool:
        """Removes the registration of a function

        Args:
            id (Union[redgrease.data.Registration, redgrease.data.ExecID, bytes, str]):
                Execution identifier for the function to unregister.

        Returns:
            bool: True if successful

        Raises:
            redis.exceptions.ResponseError:
                If the registration ID doesn't exist or if the function's reader
                doesn't support the unregister operation.
        """
        if isinstance(id, redgrease.data.Registration):
            id = id.id

        return self.redis.execute_command("RG.UNREGISTER", to_redis_type(id))


class Redis(redis.Redis):
    """Redis client class, with support for gears features.

    Behaves exactly like the redis.Redis client, but is extended with a 'gears'
    property fo executiong Gears commands.

    Attributes:
        gears (redgrease.client.Gears):
            Gears command client.
    """

    RESPONSE_CALLBACKS = {
        **redis.Redis.RESPONSE_CALLBACKS,
        **Gears.RESPONSE_CALLBACKS,
    }

    def __init__(self, *args, **kwargs):
        """Instantiate a redis client, with gears features"""
        self._gears = None
        super().__init__(*args, **kwargs)

    @property
    def gears(self) -> Gears:
        """Gears client, exposing gears commands

        Returns:
            Gears:
                Gears client
        """
        if not self._gears:
            self._gears = Gears(self)

        return self._gears



# -*- coding: utf-8 -*-
"""
Sugar classes for the various Gears Reader types.
"""
__author__ = "Anders Åström"
__contact__ = "anders@lyngon.com"
__copyright__ = "2021, Lyngon Pte. Ltd."
__licence__ = """The MIT License
Copyright © 2021 Lyngon Pte. Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the “Software”), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:

The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


from typing import Callable, Dict, Iterable, Optional

import redgrease.gears
import redgrease.sugar
import redgrease.typing
import redgrease.utils


class GearReader(redgrease.gears.PartialGearFunction[redgrease.typing.OutputRecord]):
    """Base class for the Reader sugar classes.

    Extends `redgrease.runtime.GearsBuilder' with arguments for collecting
    requirements.
    """

    def __init__(
        self,
        reader: str = redgrease.sugar.ReaderType.KeysReader,
        defaultArg: str = "*",
        desc: str = None,
        requirements: Optional[Iterable[str]] = None,
    ):
        """Gear function / process factory
        Args:
            reader (str, optional):
                Input records reader.
                Defining where the input to the gear will come from.
                One of:
                - 'KeysReader':
                - 'KeysOnlyReader':
                - 'StreamReader':
                - 'PythonReader':
                - 'ShardsReader':
                - 'CommandReader':
                Defaults to 'KeysReader'.

            defaultArg (str, optional):
                Additional arguments to the reader.
                These are usually a key's name, prefix, glob-like
                or a regular expression.
                Its use depends on the function's reader type and action.
                Defaults to '*'.

            desc (str, optional):
                An optional description.
                Defaults to None.

            requirements (Iterable[str], optional):
                Package lependencies for the gear train.
                Defaults to None.
        """
        reader_op = redgrease.gears.Reader(reader, defaultArg, desc)
        super().__init__(operation=reader_op, requirements=requirements)


class KeysReader(GearReader[redgrease.typing.Record]):
    """KeysReader is a convenience class for GearsBuilder("KeysReader", ...)"""

    def __init__(
        self,
        default_key_pattern: str = "*",
        desc: str = None,
        requirements: Optional[Iterable[str]] = None,
    ):
        """Instantiate a KeysReader partial Gear function.

        Args:
            default_key_pattern (str, optional):
                Default Redis key pattern for the keys (and its values, type) to read.
                Defaults to "*".

            desc (str, optional):
                An optional description.
                Defaults to None.

            requirements (Iterable[str], optional):
                Package lependencies for the gear train.
                Defaults to None.
        """
        super().__init__(
            reader=redgrease.sugar.ReaderType.KeysReader,
            defaultArg=default_key_pattern,
            desc=desc,
            requirements=requirements,
        )
        self.default_key_pattern = default_key_pattern

    def values(self, type=..., event=...) -> redgrease.gears.PartialGearFunction:
        """Filter out and select the values of the records only.

        Args:
            type (Union[str, Container[str]], optional):
                A single string, or a container of several strings, representing the
                Redis type(s) of keys to select.
                Valid values include: "string", "hash", "list", "set", "zset",
                "stream" or "module".
                Defaults to ... (Ellipsis), meaning any type.

            event (Union[str, Container[str]], optional):
                A single string, or a container of several strings, representing the
                Redis command or event `that .
                Defaults to ... (Ellipsis), menaing any event.

        Returns:
            PartialGearFunction:
                A new partial gear function generating the matching values.
        """
        f: redgrease.gears.PartialGearFunction = self
        if type != ... or event != ...:
            f = f.filter(redgrease.utils.dict_filter(type=type, event=event))
        return f.map(lambda record: record["value"])

    def keys(self, type=..., event=...) -> redgrease.gears.PartialGearFunction[str]:
        """Filter out and select the keys of the records only.

        Args:
            type (Union[str, Container[str]], optional):
                A single string, or a container of several strings, representing the
                Redis type(s) of keys to select.
                Valid values include: "string", "hash", "list", "set", "zset",
                "stream" or "module".
                Defaults to ... (Ellipsis), meaning any type.

            event (Union[str, Container[str]], optional):
                A single string, or a container of several strings, representing the
                Redis command or event `that .
                Defaults to ... (Ellipsis), menaing any event.

        Returns:
            PartialGearFunction[str]:
                A new partial gear function generating the matching keys.
        """
        f: redgrease.gears.PartialGearFunction = self
        if type != ... or event != ...:
            f = f.filter(redgrease.utils.dict_filter(type=type, event=event))
        return f.map(lambda record: record["key"])

    def records(
        self, type=None, event=...
    ) -> redgrease.gears.PartialGearFunction[redgrease.utils.Record]:
        """Filter out and map the records to `redgrease.utils.Record` objects.

        This provides the fields `key`, `value`, `type` and `event` as typed attributes
        on an object, instead of items in a `dict`, making it a little bit more
        pleasant to work with.

        Args:
            type (Union[str, Container[str]], optional):
                A single string, or a container of several strings, representing the
                Redis type(s) of keys to select.
                Valid values include: "string", "hash", "list", "set", "zset",
                "stream" or "module".
                Defaults to ... (Ellipsis), meaning any type.

            event (Union[str, Container[str]], optional):
                A single string, or a container of several strings, representing the
                Redis command or event `that.
                Defaults to ... (Ellipsis), menaing any event.

        Returns:
            PartialGearFunction[redgrease.utils.Record]:
                A new partial gear function generating the matching Record values.
        """
        f: redgrease.gears.PartialGearFunction = self
        if type != ... or event != ...:
            f = f.filter(redgrease.utils.dict_filter(type=type, event=event))
        return f.map(redgrease.utils.record)


class KeysOnlyReader(GearReader[str]):
    """KeysOnlyReader is a convenience class for GearsBuilder("KeysOnlyReader", ...)"""

    def __init__(
        self,
        default_key_pattern: str = "*",
        desc: str = None,
        requirements: Optional[Iterable[str]] = None,
    ):
        """Instantiate a KeysOnlyReader partial Gear function.

        Args:
            default_key_pattern (str, optional):
                Default Redis keys pattern for the keys to read.
                Defaults to "*".

            desc (str, optional):
                An optional description.
                Defaults to None.

            requirements (Iterable[str], optional):
                Package lependencies for the gear train.
                Defaults to None.
        """
        super().__init__(
            reader=redgrease.sugar.ReaderType.KeysOnlyReader,
            defaultArg=default_key_pattern,
            desc=desc,
            requirements=requirements,
        )
        self.default_key_pattern = default_key_pattern


class StreamReader(GearReader[redgrease.typing.Record]):
    """StreamReader is a convenience class for GearsBuilder("StreamReader", ...)"""

    def __init__(
        self,
        default_key_pattern: str = "*",
        desc: str = None,
        requirements: Optional[Iterable[str]] = None,
    ):
        """Instantiate a StreamReader partial Gear function.

        Args:
            default_key_pattern (str, optional):
                Default Redis keys pattern for the redis stream(s) to read.
                Defaults to "*".

            desc (str, optional):
                An optional description.
                Defaults to None.

            requirements (Iterable[str], optional):
                Package lependencies for the gear train.
                Defaults to None.
        """
        super().__init__(
            reader=redgrease.sugar.ReaderType.StreamReader,
            defaultArg=default_key_pattern,
            desc=desc,
            requirements=requirements,
        )
        self.default_key_pattern = default_key_pattern

    def values(self) -> redgrease.gears.PartialGearFunction[Dict]:
        """Select the values of the stream only.

        Returns:
            PartialGearFunction:
                A new partial gear function generating values.
        """
        return self.map(lambda record: record["value"])

    def keys(self) -> redgrease.gears.PartialGearFunction[str]:
        """Select the keys, i.e. stream names, only.

        Returns:
            PartialGearFunction[str]:
                A new partial gear function generating names.
        """
        return self.map(lambda record: record["key"])

    def records(
        self,
    ) -> redgrease.gears.PartialGearFunction[redgrease.utils.StreamRecord]:
        """Filter out and map the records to `redgrease.utils.StreamRecord` objects.

        This provides the fields `key`, `id` and `value` as typed attributes on an
        object, instead of items in a `dict`, making it a little bit more pleasant
        to work with.

        Returns:
            PartialGearFunction[redgrease.utils.Record]:
                A new partial gear function generating the matching Record values.
        """
        return self.map(redgrease.utils.stream_record)


class PythonReader(GearReader):
    """PythonReader is a convenience class for GearsBuilder("PythonReader", ...)"""

    def __init__(
        self,
        desc: str = None,
        requirements: Optional[Iterable[str]] = None,
    ):
        """Instantiate a PythonReader partial Gear function.

        Args:
            desc (str, optional):
                An optional description.
                Defaults to None.

            requirements (Iterable[str], optional):
                Package lependencies for the gear train.
                Defaults to None.
        """
        super().__init__(
            reader=redgrease.sugar.ReaderType.PythonReader,
            desc=desc,
            requirements=requirements,
        )


class ShardsIDReader(GearReader[str]):
    """ShardsIDReader is a convenience class for GearsBuilder("ShardsIDReader", ...)"""

    def __init__(
        self,
        desc: str = None,
        requirements: Optional[Iterable[str]] = None,
    ):
        """Instantiate a ShardsIDReader partial Gear function.

        Args:
            desc (str, optional):
                An optional description.
                Defaults to None.

            requirements (Iterable[str], optional):
                Package lependencies for the gear train.
                Defaults to None.
        """
        super().__init__(
            reader=redgrease.sugar.ReaderType.ShardsIDReader,
            desc=desc,
            requirements=requirements,
        )


class CommandReader(GearReader):
    """CommandReader is a convenience class for GearsBuilder("CommandReader", ...)"""

    def __init__(
        self,
        desc: str = None,
        requirements: Optional[Iterable[str]] = None,
    ):
        """Instantiate a CommandReader partial Gear function.

        Args:
            desc (str, optional):
                An optional description.
                Defaults to None.

            requirements (Iterable[str], optional):
                Package lependencies for the gear train.
                Defaults to None.
        """
        super().__init__(
            reader=redgrease.sugar.ReaderType.CommandReader,
            desc=desc,
            requirements=requirements,
        )

    def args(self, max_count: int = None) -> redgrease.gears.PartialGearFunction:
        """Ignore the trigger name and take only the arguments following the trigger.

        Args:
            max_count (int, optional):
                Maximum number of args to take. Any additional args will be truncated.
                Defaults to None.

        Returns:
            redgrease.gears.PartialGearFunction:
                A new partial gear function only generating the trigger arguments.
        """
        if max_count:
            max_count += 1
        return self.map(lambda x: x[1:max_count])

    def apply(
        self, fun: Callable[..., redgrease.typing.OutputRecord]
    ) -> redgrease.gears.PartialGearFunction[redgrease.typing.OutputRecord]:
        """Apply a function to the trigger arguments.

        Args:
            fun (Callable[..., redgrease.typing.OutputRecord]):
                The function to call with the trigger arguments.

        Returns:
            redgrease.gears.PartialGearFunction[redgrease.typing.OutputRecord]:
                A new partial gear function generating the results of the function.
        """
        return self.map(lambda args: fun(*args[1:]))



from typing import Callable, Dict

import rediscluster
import rediscluster.exceptions

import redgrease.client
import redgrease.data
import redgrease.utils


class RedisCluster(rediscluster.RedisCluster):
    """RedisCluster client class, with support for gears features

    Behaves exactly like the rediscluster.RedisCluster client, but is extended with
    a 'gears' property fo executiong Gears commands.

    Attributes:
        gears (redgrease.client.Gears):
            Gears command client.
    """

    # States how target node is selected for cluster commands:
    # - blocked : command is not allowed - Raises a RedisClusterException
    # - random : excuted on one randomly selected node
    # - all-masters : executed on all master node
    # - all-nodes : executed on all nodes
    # - slot-id : executed on the node defined by the second argument
    NODES_FLAGS = {
        **rediscluster.RedisCluster.NODES_FLAGS,
        **{
            "RG.INFOCLUSTER": "random",
            "RG.PYSTATS": "all-nodes",
            "RG.PYDUMPREQS": "random",
            "RG.REFRESHCLUSTER": "all-nodes",
        },
    }

    # Not to be confused with redis.Redis.RESPONSE_CALLBACKS
    # RESULT_CALLBACKS is special to rediscluster.RedisCluster.
    # It decides how results of commands defined in `NODES_FLAGS` are aggregated into
    # a final response, **after** redis.Redis.RESPONSE_CALLBACKS as been applied to
    # each response individually.
    RESULT_CALLBACKS = {
        **rediscluster.RedisCluster.RESULT_CALLBACKS,
        **{
            "RG.INFOCLUSTER": lambda _, res: next(iter(res.values())),
            "RG.PYSTATS": lambda _, res: res,
            "RG.PYDUMPREQS": lambda _, res: next(iter(res.values())),
            "RG.REFRESHCLUSTER": lambda _, res: all(res.values()),
        },
    }

    RESPONSE_CALLBACKS: Dict[str, Callable] = {
        **rediscluster.RedisCluster.RESPONSE_CALLBACKS,
        **redgrease.client.Gears.RESPONSE_CALLBACKS,
    }

    def __init__(self, *args, **kwargs):
        """Instantiate a redis cluster client, with gears features"""
        self._gears = None
        self.connection = None
        super().__init__(*args, **kwargs)

    @property
    def gears(self) -> redgrease.client.Gears:
        """Gears client, exposing gears commands

        Returns:
            Gears:
                Gears client
        """
        if not self._gears:
            self._gears = redgrease.client.Gears(self)

        return self._gears


def RedisGears(*args, **kwargs):
    try:
        return RedisCluster(*args, **kwargs)

    except (AttributeError, rediscluster.exceptions.RedisClusterException):
        return redgrease.client.Redis(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
Handling of events that come in bursts, without unecessary duplicate handling.

In RedGrease this is used in the cli / loader, when watching a directory and
'hot reloading' of scripts when a change is detedcted.
To avoid scripts being loaded unesssearily often while being modified.

This module is however general enough to be used for other situatins where events occur
in bursts.
"""
__author__ = "Anders Åström"
__contact__ = "anders@lyngon.com"
__copyright__ = "2021, Lyngon Pte. Ltd."
__licence__ = """The MIT License
Copyright © 2021 Lyngon Pte. Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the “Software”), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:

The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


from threading import Timer
from typing import Callable, Dict, Hashable, Optional


def nop(*args, **kwargs):
    """No Operation: Do nothing, regardless what is passed as arguments"""
    pass


class ResettableTimer:
    """Simple reresttable timer that will call a function when the timer elapses.
    Both the timer, as well as the function and its arguments can be reset
    when/if the timer is reset.
    """

    def __init__(self):
        """Create timer"""
        self.handler = nop
        self.handler_args = []
        self.handler_kwargs = {}
        self.timer = None

    def __del__(self):
        """Destructor. Cancels timer"""
        self.cancel()

    def cancel(self):
        """Cancel the timer (if set) so that it wont elapse"""
        if self.timer:
            try:
                self.timer.cancel()
            except Exception:
                pass
        self.timer = None

    def set(
        self,
        timeout: Optional[float],
        handler: Callable[..., None],
        *handler_args,
        **handler_kwargs
    ):
        """Reset the timer and callback

        Args:
            timeout (float):
                New timeout, in seconds.

            handler (Callable[..., None]):
                Handler function that will be called after `timeout` duration from
                **last** received event.
                Any new events received before will 'reset' the interval.

            *handler_args (Any):
                Any additional arguments for the handler, when triggered.

            **handler_kwargs (Any):
                Any additional keword arguments for the handler, when triggered.
        """
        self.cancel()
        self.handler = handler
        self.handler_args = handler_args
        self.handler_kwargs = handler_kwargs
        self.timer = Timer(timeout if timeout else 1.0, self.trigger)
        self.timer.start()

    def trigger(self):
        """Trigger the handler immediately. Will cancel the timer."""
        self.cancel()
        if self.handler:
            self.handler(*self.handler_args, **self.handler_kwargs)


class HysteresisHandlerIndex:
    """
    Index of multiple signal handlers for different signals, triggering each handler
    only when its' corresponding signal has not been registered for a set amount
    of time.

    This is useful for holding back some event handler as long as new events
    are still occuring, triggering only when the events stop.
    For example loading a file when there hasnt been any modifications to it
    for a period of time
    """

    def __init__(self, hysteresis_duration: Optional[float] = 1.0):
        """Initialize the Index.

        Args:
            hysteresis_duration (float, optional):
                The required number of seconds without any signals for a given handler
                for it to trigger.
            Defaults to 1.0 second.
        """
        self.hysteresis_duration = hysteresis_duration
        self.handlers: Dict[Hashable, ResettableTimer] = {}

    def signal(
        self,
        handler_id: Hashable,
        handler_function: Callable[..., None],
        *handler_args,
        **handler_kwargs
    ):
        """Register a new signal for a specific handler.

        Args:
            handler_id ([Hashable]):
                Id / name of the handler / signal / event.

            handler_function (Callable[..., None]):
                Signal handler function.

            *handler_args (Any):
                Any additional arguments for the handler, when triggered.

            **handler_kwargs (Any):
                Any additional keword arguments for the handler, when triggered.
        """
        if handler_id not in self.handlers:
            self.handlers[handler_id] = ResettableTimer()

        self.handlers[handler_id].set(
            self.hysteresis_duration, handler_function, *handler_args, **handler_kwargs
        )



# -*- coding: utf-8 -*-
"""
Module for loading Gears script files into a Redis Gears cluster, as well as
continiously monitor the files or directories for changes, and in that case reload the
modified/created scripts into the cluster.
"""
__author__ = "Anders Åström"
__contact__ = "anders@lyngon.com"
__copyright__ = "2021, Lyngon Pte. Ltd."
__licence__ = """The MIT License
Copyright © 2021 Lyngon Pte. Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the “Software”), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:

The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import logging
import re
from datetime import datetime
from fnmatch import fnmatch
from os.path import isfile
from pathlib import Path
from typing import List, Union

from redis.exceptions import RedisError, ResponseError
from watchdog.events import PatternMatchingEventHandler
from watchdog.observers import Observer

from redgrease import RedisGears, formatting, hysteresis, requirements

log = logging.getLogger(__name__)

default_index_prefix = "/redgrease/scripts"
"""The defaut prefix for the Redis keys that hold the Loader index data"""

default_script_pattern = "*.py"
"""The default pattern for scripts to monitor and load."""

default_requirements_pattern = "*requirements*.txt"
"""The default pattern for requirements files to monitor and load."""

default_unblocking_pattern = "unblock"
"""The default pattern (within script files) that indicate 'unblocking' load"""

default_ignore_patterns: List[str] = []
"""Default list of file patterns to ignore"""

# Regex type changed
# from re.SRE_Pattern in Python 3.6
# to re.Pattern in Python3 3.7
_regex_type = type(re.compile(""))


def fail(exception, *messages):
    """Convenience function for raising exceptions

    Args:
        exception ([Exception]):
            Exception to raise.

        *messages (str):
            Any number of messages.
            Will be separated by full stop/period (.).

    Raises:
        exception: Exception
    """
    error_message = ". ".join(messages)
    log.error(error_message)

    raise exception(error_message)


class GearsLoader:
    """Monitors and loads one or more script files into Redis Gear cluster.

    Note that the loader will use the Redis cluster to store some indexing data
    for its own operation.
    Most notably information about which registration id has bee assigned to any
    'registered' script files.
    """

    def __init__(
        self,
        script_pattern: str = None,
        requirements_pattern: str = None,
        unblocking_pattern=None,
        ignore_patterns: str = None,
        index_prefix: str = None,
        server: str = "localhost",
        port: int = 6379,
        observe: float = None,
        **redis_kwargs,
    ):
        """Instatiate a Gears Loader

        Args:
            script_pattern (str, optional):
                A file pattern the files to monitor. Supports globbing.
                E.g. `scripts/*.py`.
                Defaults to None.

            requirements_pattern (str, optional):
                A file pattern for any requirement files defining the requirements of
                the scripts monitored. Supports globbing.
                E.g. `scripts/*requirements*.txt`.
                Defaults to None.

            unblocking_pattern ([type], optional):
                Pattern within script files that determine if the files should be
                loaded as "unblocking".
                E.g. `unblock`.
                Defaults to None.

            ignore_patterns (str, optional):
                Patterns for files to ignore.
                Defaults to None.

            index_prefix (str, optional):
                Prefix for the keys in Redis used by the GearLoader index.
                Defaults to None.

            server (str, optional):
                Server / host name or IP.
                Defaults to "localhost".

            port (int, optional):
                Server / host port number.
                Defaults to 6379.

            observe (float, optional):
                If `True`, continously observe the files, and reload if there are any
                changes.
                If `False`, the files will only be loaded when added to the Loader.
                Defaults to None.
        """
        self.directories: List[Path] = []

        self.script_pattern = (
            script_pattern if script_pattern else default_script_pattern
        )

        self.requirements_pattern = (
            requirements_pattern
            if requirements_pattern
            else default_requirements_pattern
        )

        if unblocking_pattern is None:
            unblocking_pattern = default_unblocking_pattern
        if isinstance(unblocking_pattern, str):
            unblocking_pattern = re.compile(re.escape(unblocking_pattern))
        if not isinstance(unblocking_pattern, _regex_type):
            fail(ValueError, f"Invalid unblocking pattern: {unblocking_pattern}")
        self.unblocking_pattern = unblocking_pattern

        self.ignore_patterns = (
            ignore_patterns if ignore_patterns else default_ignore_patterns
        )

        self.index_prefix = index_prefix if index_prefix else default_index_prefix

        self.redis = RedisGears(host=server, port=port, **redis_kwargs)

        self.observer = Observer() if observe else None

        self.file_index = hysteresis.HysteresisHandlerIndex(hysteresis_duration=observe)

        # Create the file change event listener
        self.event_handler = (
            PatternMatchingEventHandler(
                patterns=[self.script_pattern, self.requirements_pattern],
                ignore_patterns=self.ignore_patterns,
                ignore_directories=False,
                case_sensitive=False,
            )
            if self.observer
            else None
        )

        if self.event_handler:
            self.event_handler.on_deleted = self.on_deleted
            self.event_handler.on_modified = self.on_modified
            self.event_handler.on_moved = self.on_moved

    def __del__(self):
        try:
            self.redis.close()
        except RedisError as ex:
            log.warn(f"Error while closing redis: {ex}")

        try:
            self.stop()
        except Exception as ex:
            log.warn(f"Error stopping observer: {ex}")
            raise

    def add_directory(
        self, directory: Union[Path, str], recursive: bool = False
    ) -> None:
        """Add a directory to the Loader

        Args:
            directory (Union[Path, str]):
                Directory of scripts to load.

            recursive (bool, optional):
                Recursively scan sub-directories.
                Defaults to False.
        """
        if not isinstance(directory, Path):
            directory = Path(str(directory))

        log.info(f"Adding event handlers for {directory}")

        # select either recursive or non-recursive glob function for the dir
        globber = directory.rglob if recursive else directory.glob

        # find and run/register all requirements files in the watch directories
        for requirements_file_path in globber(self.requirements_pattern):
            self.update_dependencies(requirements_file_path)

        # find and run/register all script files in the watch directory
        for script_file_path in globber(self.script_pattern):
            self.register_script(script_file_path)

        if self.observer is not None:
            self.observer.schedule(self.event_handler, directory, recursive)

        self.directories.append(directory)

    # TODO: Should call pyexecute with the file path directly
    # TODO: Handle multiple registrations per script.
    def register_script(self, script_path):
        """Execute / Register a gear script in redis using 'RG.PYEXECUTE'

            Note: paths that maches the 'unblocking-pattern', will be executed
            with the 'UNBLOCKING' modifier.

        Args:
            script_path (str):
                Path to the script.

        """
        log.debug(f"Registering script '{script_path}'")
        general_failure_msg = f"Unable to register script file '{script_path}'"

        if not isfile(script_path):
            fail(FileNotFoundError, general_failure_msg, "File not found")

        with open(script_path) as script_file:
            script_content = script_file.read()

        if not script_content:
            fail(IOError, general_failure_msg, "File is empty")

        # Unregister script if already present
        self.unregister_script(script_path)

        unblocking = self.unblocking_pattern.search(str(script_path)) is not None

        try:
            log.debug(
                f"Running/registering Gear script '{script_path}' " "on Redis server"
            )

            # This is a quite unsafe way of checking for registrations
            # Probabls Ok for dev situations in non-shared environments
            pre_reg = self.redis.gears.dumpregistrations()
            exec_res = self.redis.gears.pyexecute(script_content, unblocking=unblocking)
            ret = f"with return code '{exec_res}'."
            post_reg = self.redis.gears.dumpregistrations()
            pre_reg = set([reg.id for reg in pre_reg])
            log.debug(f"Pre regs: {list(pre_reg)}")
            post_reg = set([reg.id for reg in post_reg])
            log.debug(f"Post regs: {list(post_reg)}")
            diff_reg: set = post_reg - pre_reg
            if len(diff_reg) > 0:
                reg_id = diff_reg.pop()
                self.redis.hset(
                    f"{self.index_prefix}{script_path}",
                    mapping={
                        "registration_id": str(reg_id),
                        "last_updated": datetime.utcnow().strftime(
                            formatting.iso8601_datefmt
                        ),
                    },
                )
                log.debug(f"Script '{script_path}' registered as '{reg_id}' {ret}")
            else:
                log.debug(f"Script '{script_path}' executed {ret}")

            if len(diff_reg) > 0:
                log.warning(
                    "Multiple registrations occured? "
                    "Index migth be corrupt. "
                    "Is this a shared environment?"
                )

        except RedisError as ex:
            log.error(f"Something went wrong: {ex}")

    # Actions
    # TODO: Should call pyexecute with the file path directly
    # TODO: Handle multiple registrations per script.
    def unregister_script(self, script_path):
        """Check if a given script is registered, and if so,
        unregister it using 'RG.UNREGISTER'

        Args:
            script_path (str):
                Script path
        """
        log.debug(f"Unregistering script: '{script_path}'")
        reg_key = f"{self.index_prefix}{script_path}"
        reg_id = self.redis.hget(reg_key, "registration_id")
        if reg_id is not None:
            log.debug(
                f"Removing registration for script '{script_path}' "
                f"with id '{reg_id}'"
            )
            try:
                self.redis.gears.unregister(reg_id)
            except ResponseError as err:
                log.warn(
                    "Unregistration failed. "
                    "Index migth be corrupt. "
                    "Is this a shared environment?"
                )
                log.error({err})
        self.redis.delete(reg_key)

    def update_dependencies(self, requirements_file_path):
        """Update (add only) package dependencies on the Redis instance

        Args:
            requirements_file_path (str):
                File path of 'requirements.txt' file.
        """
        log.debug(f"Updating dependecies as per '{requirements_file_path}'")
        try:
            requirements_set = requirements.read_requirements(requirements_file_path)
            log.debug(f"Requirements to load: {', '.join(requirements_set)}")
            self.redis.gears.pyexecute(requirements=requirements_set)
        except Exception as ex:
            log.error(f"Something went wrong: {ex}")

    # File Event Handling
    def on_deleted(self, event):
        """Watchdog event handler for events signalling that a
        script or requirement file has been deleted.
        Script files are unregistered (if present) and re-run,
        after applying some hystersis.
        Removed requirement files does not remove any installed packages.

        Args:
            event (watchdog.FileDeletedEvent):
                Event data for deleted file
        """
        file = event.src_path
        if fnmatch(file, self.script_pattern):
            log.debug(
                f"Gears script '{file}' deleted. "
                "Scheduling unregistration of script."
            )
            # Apply hysteresis in case additional events shortly follow,
            # before we actually unregister
            self.file_index.signal(file, self.unregister_script, file)
        elif fnmatch(file, self.requirements_pattern):
            log.info(
                f"Gears requirements file '{file}' deleted. "
                "Requirement removal not Implemented. "
                "Ignoring."
            )
        else:
            log.warn(f"Unknown file type '{file}' deleted. " "Ignoring.")

    def on_modified(self, event):
        """Watchdog event handler for events signalling that a
        script or requirement file has been modified.

        Args:
            event (watchdog.FileModifiedEvent):
                Event data for modified file
        """
        file = event.src_path
        if fnmatch(file, self.script_pattern):
            log.debug(f"Gears script '{file}' modified. Regestering script.")
            # Apply hysteresis in case additional events shortly follow,
            # before we actually (re-)run/register the script.
            self.file_index.signal(file, self.register_script, file)
        elif fnmatch(event.src_path, self.requirements_pattern):
            log.debug(
                f"Gears requirements file '{file}' modified. " "Updating dependencies."
            )
            # Apply hysteresis in case additional events shortly follow,
            # before we actually update requirements.
            self.file_index.signal(file, self.update_dependencies, file)
        else:
            log.warn(f"Unknown file type '{event.src_path}' modified. Ignoring.")

    def on_moved(self, event):
        """Watchdog event handler for events signalling that a
        script or requirement file has been moved.
        Script files are unregistered (if present) and re-run,
        after applying some hystersis.
        Removed requirement files does not remove any installed packages.

        Args:
            event (watchdog.FileMovedEvent):
                Event data for moved file.
        """
        old_file = event.src_path
        new_file = event.dest_path
        # Handle, i.e. unregister, the old / source file.
        if fnmatch(old_file, self.script_pattern):
            log.debug(
                f"Gears script '{old_file}' moved to '{new_file}'. "
                "Unregistering old script."
            )
            # Apply hysteresis in case additional events shortly follow,
            # before we actually unregister the script.
            self.file_index.signal(old_file, self.unregister_script, old_file)
        elif fnmatch(old_file, self.requirements_pattern):
            log.debug(
                f"Gears requirements file '{old_file}' moved to '{new_file}'. "
                "Requirement removal not Implemented. "
                "Ignoring."
            )

        # TODO: Check that the new file is in the watch directory
        # Handle, i.e. run/register, the new / destination file
        if fnmatch(new_file, self.script_pattern):
            log.debug(
                f"File '{old_file}' moved to Gears script '{new_file}'. "
                "Registering new script."
            )
            # Apply hysteresis in case additional events shortly follow,
            # before we actually (re-)run/register the script
            self.file_index.signal(new_file, self.register_script, new_file)
        elif fnmatch(new_file, self.requirements_pattern):
            log.debug(
                f"File '{old_file}' moved to requirements file '{new_file}'. "
                "Updating dependencies"
            )
            # Apply hysteresis in case additional events shortly follow,
            # before we actually update requirements.
            self.file_index.signal(new_file, self.update_dependencies, new_file)
        else:
            log.warn(
                f"File '{old_file}' moved to unknonwn type '{new_file}'. " "Ignoring."
            )

    def start(self):
        """Start Loader monitoring."""
        log.info("Starting RedGrease directory watcher ")
        if self.observer:
            self.observer.start()

    def stop(self):
        """Stop Loader monitoring."""
        if self.observer:
            self.observer.stop()



# -*- coding: utf-8 -*-
"""
Redis Gears Config
"""
__author__ = "Anders Åström"
__contact__ = "anders@lyngon.com"
__copyright__ = "2021, Lyngon Pte. Ltd."
__licence__ = """The MIT License
Copyright © 2021 Lyngon Pte. Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the “Software”), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:

The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import logging
from typing import Any, Dict, Union

import redis

from redgrease.typing import Constructor
from redgrease.utils import safe_str, to_list

log = logging.getLogger(__name__)


class Config:
    """Redis Gears Config"""

    ValueTypes: Dict[str, Constructor] = {
        "MaxExecutions": int,
        "MaxExecutionsPerRegistration": int,
        "ProfileExecutions": bool,
        "PythonAttemptTraceback": bool,
        "DownloadDeps": bool,
        "DependenciesUrl": safe_str,
        "DependenciesSha256": safe_str,
        "PythonInstallationDir": safe_str,
        "CreateVenv": bool,
        "ExecutionThreads": int,
        "ExecutionMaxIdleTime": int,
        "PythonInstallReqMaxIdleTime": int,
        "SendMsgRetries": int,
    }
    """Mapping from config name to its corresponding value type, transformer
    or constructor.
    """

    __slots__ = "redis"

    def __init__(self, redis: redis.Redis):
        """Instantiate a Redis Gears Confg object

        Args:
            redis (redis.Redis):
                Redis client / connection for underlying communication.
        """
        self.redis = redis

    def get(self, *config_option: Union[bytes, str]) -> Dict[Union[bytes, str], Any]:
        """Get the value of one or more built-in configuration or
        a user-defined options.

        Args:
            *config_option (Union[bytes, str]):
                One or more names/key of configurations to get.

        Returns:
            Dict[Union[bytes, str], Any]:
                Dict of the requested config options mapped to their corresponding
                values.
        """
        if config_option:
            conf = [safe_str(c) for c in config_option]
        else:
            conf = list(Config.ValueTypes.keys())

        return self.redis.execute_command("RG.CONFIGGET", *conf, keys=conf)

    def set(self, config_dict=None, **config_setting) -> bool:
        """Set a value of one ore more built-in configuration or
        a user-defined options.

        This function offers two methods of providing the keys and values to set;
        either as a mapping/dict or b key-word arguments.

        Args:
            config_dict  (Mapping[str, Any]): Mapping / dict of config values to set.

            **config_setting (Any): Key-word arguments to set as config values.

        Returns:
            bool:
                `True` if all was successful, `False` oterwise
        """
        settings = {}
        if config_dict:
            settings.update(config_dict)

        if config_setting:
            settings.update(config_setting)

        return self.redis.execute_command("RG.CONFIGSET", *to_list(settings))

    def get_single(self, config_option: Union[bytes, str]):
        """Get a single config value.

        Args:
            config_option (str):
                Name of the config to get.

        Returns:
            Any:
                The value of the config option.
        """
        return self.get(config_option)[config_option]

    @property
    def MaxExecutions(self) -> int:
        """Get the current value for the `MaxExecutions` config option.

        The MaxExecutions configuration option controls the maximum number of
        executions that will be saved in the executions list.
        Once this threshold value is reached, older executions will be deleted from
        the list by order of their creation (FIFO).

        Only executions that had finished (e.g. the 'done' or 'aborted' status ) are
        deleted.
        """
        return self.get_single("MaxExecutions")

    @MaxExecutions.setter
    def MaxExecutions(self, value: int) -> None:
        self.set(MaxExecutions=value)

    @property
    def MaxExecutionsPerRegistration(self) -> int:
        """Get the current value for the `MaxExecutionsPerRegistration` config option.

        The MaxExecutionsPerRegistration configuration option controls the maximum
        number of executions that are saved in the list per registration.
        Once this threshold value is reached, older executions for that registration
        will be deleted from the list by order of their creation (FIFO).

        Only executions that had finished (e.g. the 'done' or 'aborted' status ) are
        deleted.
        """
        return self.get_single("MaxExecutionsPerRegistration")

    @MaxExecutionsPerRegistration.setter
    def MaxExecutionsPerRegistration(self, value: int) -> None:
        self.set(MaxExecutionsPerRegistration=value)

    @property
    def ProfileExecutions(self) -> bool:
        """Get the current value for the `ProfileExecutions` config option.

        The ProfileExecutions configuration option controls whether executions are
        profiled.

        **Note: Profiling impacts performance**
        Profiling requires reading the server's clock, which is a costly operation in
        terms of performance. Execution profiling is recommended only for debugging
        purposes and should be disabled in production.
        """
        return self.get_single("ProfileExecutions")

    @ProfileExecutions.setter
    def ProfileExecutions(self, value: bool) -> None:
        self.set(ProfileExecutions=value)

    @property
    def PythonAttemptTraceback(self) -> bool:
        """Get the current value for the `PythonAttemptTraceback` config option.

        The PythonAttemptTraceback configuration option controls whether the engine
        tries producing stack traces for Python runtime errors.
        """
        return self.get_single("PythonAttemptTraceback")

    @PythonAttemptTraceback.setter
    def PythonAttemptTraceback(self, value: bool) -> None:
        self.set(PythonAttemptTraceback=value)

    @property
    def DownloadDeps(self) -> bool:
        """Get the current value for the `DownloadDeps` config option.

        The DownloadDeps configuration option controls whether or not RedisGears will
        attempt to download missing Python dependencies.
        """
        return self.get_single("DownloadDeps")

    # @DownloadDeps.setter
    # def DownloadDeps(self, value):
    #     self.set(DownloadDeps=value)

    @property
    def DependenciesUrl(self) -> str:
        """Get the current value for the `DependenciesUrl` config option.

        The DependenciesUrl configuration option controls the location from which
        RedisGears tries to download its Python dependencies.
        """
        return self.get_single("DependenciesUrl")

    # @DependenciesUrl.setter
    # def DependenciesUrl(self, value):
    #     self.set(DependenciesUrl=value)

    @property
    def DependenciesSha256(self) -> str:
        """Get the current value for the `DependenciesSha256` config option.

        The DependenciesSha256 configuration option specifies the SHA265 hash value
        of the Python dependencies. This value is verified after the dependencies have
        been downloaded and will stop the server's startup in case of a mismatch.
        """
        return self.get_single("DependenciesSha256")

    # @DependenciesSha256.setter
    # def DependenciesSha256(self, value):
    #     self.set(DependenciesSha256=value)

    @property
    def PythonInstallationDir(self):
        """Get the current value for the `PythonInstallationDir` config option.

        The PythonInstallationDir configuration option specifies the path for
        RedisGears' Python dependencies.
        """
        return self.get_single("PythonInstallationDir")

    # @PythonInstallationDir.setter
    # def PythonInstallationDir(self, value):
    #     self.set(PythonInstallationDir=value)

    @property
    def CreateVenv(self):
        """Get the current value for the `CreateVenv` config option.

        The CreateVenv configuration option controls whether the engine will create a
        virtual Python environment.
        """
        return self.get_single("CreateVenv")

    # @CreateVenv.setter
    # def CreateVenv(self, value):
    #     self.set(CreateVenv=value)

    @property
    def ExecutionThreads(self):
        """Get the current value for the `ExecutionThreads` config option.

        The ExecutionThreads configuration option controls the number of threads that
        will run executions.
        """
        return self.get_single("ExecutionThreads")

    # @ExecutionThreads.setter
    # def ExecutionThreads(self, value):
    #     self.set(ExecutionThreads=value)

    @property
    def ExecutionMaxIdleTime(self):
        """Get the current value for the `ExecutionMaxIdleTime` config option.

        The ExecutionMaxIdleTime configuration option controls the maximal amount of
        idle time (in milliseconds) before execution is aborted. Idle time means no
        progress is made by the execution.

        The main reason for idle time is an execution that's blocked on waiting for
        records from another shard that had failed (i.e. crashed).
        In that case, the execution will be aborted after the specified time limit.
        The idle timer is reset once the execution starts progressing again.
        """
        return self.get_single("ExecutionMaxIdleTime")

    @ExecutionMaxIdleTime.setter
    def ExecutionMaxIdleTime(self, value):
        self.set(ExecutionMaxIdleTime=value)

    @property
    def PythonInstallReqMaxIdleTime(self):
        """Get the current value for the `PythonInstallReqMaxIdleTime` config option.

        The PythonInstallReqMaxIdleTime configuration option controls the maximal
        amount of idle time (in milliseconds) before Python's requirements installation
        is aborted. Idle time means that the installation makes no progress.

        The main reason for idle time is the same as for ExecutionMaxIdleTime .
        """
        return self.get_single("PythonInstallReqMaxIdleTime")

    @PythonInstallReqMaxIdleTime.setter
    def PythonInstallReqMaxIdleTime(self, value):
        self.set(PythonInstallReqMaxIdleTime=value)

    @property
    def SendMsgRetries(self):
        """Get the current value for the `SendMsgRetries` config option.
        The SendMsgRetries configuration option controls the maximum number of retries
        for sending a message between RedisGears' shards. When a message is sent and
        the shard disconnects before acknowledging it, or when it returns an error,
        the message will be resent until this threshold is met.

        Setting the value to 0 means unlimited retries.
        """
        return self.get_single("SendMsgRetries")

    @SendMsgRetries.setter
    def SendMsgRetries(self, value):
        self.set(SendMsgRetries=value)



# -*- coding: utf-8 -*-
"""
GearsFunction and Operation definintions
"""
__author__ = "Anders Åström"
__contact__ = "anders@lyngon.com"
__copyright__ = "2021, Lyngon Pte. Ltd."
__licence__ = """The MIT License
Copyright © 2021 Lyngon Pte. Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the “Software”), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:

The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


from typing import Any, Dict, Generic, Hashable, Iterable, Optional, Type, TypeVar

import redgrease
import redgrease.sugar as sugar
import redgrease.typing as optype

T = TypeVar("T")


################################################################################
#                                Operations                                    #
################################################################################


class Operation:
    """Abstract base class for Gear function operations.

    Operations are the building block of RedisGears functions.
    Different operation types can be used to achieve a variety of results to
    meet various data processing needs.

    Operations can have zero or more arguments that control their operation.
    Depending on the operation's type arguments may be language-native data types
    and function callbacks.

    Attributes:
        kwargs (dict):
            Any left-over keyword arguments not explicitly consumed by the operation.
    """

    def __init__(self, **kwargs):
        """Instantiate a Operation with left-over args"""
        self.kwargs = kwargs

    def add_to(self, function):
        """Placeholder method for adding the operation to the end of a Gear function.
        This method must be implemented in each subclass, and will throw a
        ``NotImplementedException`` if called directly on the `Operation` superclass.

        Args:
            function (Union[Type, PartialGearFunction]):
                The partial gear function to append this operation to.

                If, and only if, the operation is a reader function (always and only
                the first operation in any Gear function), then the `function`
                argument must instead be a GearsBuilder class/type.

        Raises:
            NotImplementedError:
                If invoked on the `Operation` superclass, and,or not implemented in the
                subclass.
        """
        raise NotImplementedError(
            "Builder Operation has not implemented the `add_to` method: "
            f"'{self.__class__.__name__}'"
        )


class Nop(Operation):
    """No Operation.

    This Operation does nothing.
    """

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Returns the function, unmodified.

        Args:
            function (Union[Type, PartialGearFunction]):
                The partial gear function to append this operation to.

                If, and only if, the operation is a reader function (always and only
                the first operation in any Gear function), then the `function`
                argument must instead be a GearsBuilder class/type.

        Returns:
            PartialGearFunction:
                The function unmodified.
        """
        return function


class Reader(Operation):
    """Reader operation

    The Reader operation is always and only the first operation of any GearFunction.

    It defines which reader type to use and its arguments.

    Attributes:
        reader (str):
            The type of reader (https://oss.redislabs.com/redisgears/readers.html)
                - 'KeysReader'
                - 'KeysOnlyReader'
                - 'StreamReader'
                - 'PythonReader'
                - 'ShardsIDReader'
                - 'CommandReader'

        defaultArg (str):
            Argument that the reader may need. These are usually a key's name, prefix,
            glob-like or a regular expression. Its use depends on the function's reader
            type and action.

        desc (str):
            Function description.
    """

    def __init__(
        self,
        reader: str = sugar.ReaderType.KeysReader,
        defaultArg: str = "*",
        desc: Optional[str] = None,
        **kwargs,
    ) -> None:
        """Instantiate Reader operation

        Args:
            reader (str, optional):
                Reader type.
                Defaults to sugar.ReaderType.KeysReader.

            defaultArg (str, optional):
                Reader default arguments.
                Defaults to "*".

            desc (Optional[str], optional):
                Function description.
                Defaults to None.
        """
        super().__init__(**kwargs)
        self.reader = reader
        self.defaultArg = defaultArg
        self.desc = desc

    def add_to(self, builder: Type) -> "PartialGearFunction":
        """Create a new gear function based on this reader information.

        Args:
            builder (Type):
                GearsBuilder class. Defines the constuctor to use to create the
                Gear function.

        Returns:
            PartialGearFunction:
                Returns a minimal partial gear function, consisting only of the
                reader.
        """
        return builder(self.reader, self.defaultArg, self.desc, **self.kwargs)


class Run(Operation):
    """Run action

    The Run action runs a Gear function as a batch.
    The function is executed once and exits once the data is exhausted by its reader.

    The Run action can only be the last operation of any GearFunction, and it
    effectivesly 'closes' it to further operations.

    Attributes:
        arg (str, optional):
            Argument that's passed to the reader, overriding its defaultArg.
            It means the following:

                - A glob-like pattern for the KeysReader and KeysOnlyReader readers.
                - A key name for the StreamReader reader.
                - A Python generator for the PythonReader reader.

        convertToStr (bool):
            When `True`, adds a map operation to the flow's end that stringifies
            records.

        collect (bool):
            When `True`, adds a collect operation to flow's end.
    """

    def __init__(
        self,
        arg: Optional[str] = None,
        convertToStr: bool = True,
        collect: bool = True,
        **kwargs,
    ) -> None:
        """Instantiate a Run action

        Args:
            arg (Optional[str], optional):
                Optional argument that's passed to the reader, overriding its
                defaultArg.
                It means the following:

                    - A glob-like pattern for the KeysReader and KeysOnlyReader readers.
                    - A key name for the StreamReader reader.
                    - A Python generator for the PythonReader reader.

                Defaults to None.

            convertToStr (bool, optional):
                When `True`, adds a map operation to the flow's end that stringifies
                records.
                Defaults to True.

            collect (bool, optional):
                When `True`, adds a collect operation to flow's end.
                Defaults to True.
        """
        super().__init__(**kwargs)
        self.arg = arg
        self.convertToStr = convertToStr
        self.collect = collect

    def add_to(self, function: "PartialGearFunction") -> "ClosedGearFunction":
        """Closes a Gear function with the Run action.

        Args:
            function (PartialGearFunction):
                The partial Gear function to close with the run action.

        Returns:
            ClosedGearFunction:
                A closed Gear batch function that is ready to run on a Gears cluster.
        """
        import cloudpickle

        return function.map(lambda x: cloudpickle.dumps(x, protocol=4)).run(
            self.arg, False, self.collect, **self.kwargs
        )


class Register(Operation):
    """Register action

    The Register action registers a function as an event handler.
    The function is executed each time an event arrives.
    Each time it is executed, the function operates on the event's data and once done
    it is suspended until its future invocations by new events.

    The Register action can only be the last operation of any GearFunction, and it
    effectivesly 'closes' it to further operations.

    Attributes:
        prefix (str):
            Key prefix pattern to match on.
            Not relevant for 'CommandReader' readers (see 'trigger').

        convertToStr (bool):
            When `True`, adds a map operation to the flow's end that stringifies
            records.

        collect (bool):
            When `True`, adds a collect operation to flow's end.

        mode (str):
            The execution mode of the triggered function.

        onRegisered (Callable):
            A function callback that's called on each shard upon function registration.s
    """

    def __init__(
        self,
        prefix: str = "*",
        convertToStr: bool = True,
        collect: bool = True,
        mode: str = sugar.TriggerMode.Async,
        onRegistered: "optype.Callback" = None,
        **kwargs,
    ) -> None:
        """Instantiate a Register action

        Args:
            prefix (str, optional):
                Key prefix pattern to match on.
                Not relevant for 'CommandReader' readers (see 'trigger').
                Defaults to '*'.

            convertToStr (bool, optional):
                When `True` adds a map operation to the flow's end that stringifies
                records.
                Defaults to True.

            collect (bool, optional):
                When True adds a collect operation to flow's end.
                Defaults to False.

            mode (str, optional):
                The execution mode of the function.
                Can be one of::

                - 'async' : Execution will be asynchronous across the entire cluster.

                - 'async_local' : Execution will be asynchronous and restricted to
                    the handling shard.

                - 'sync' : Execution will be synchronous and local

                Defaults to `redgrease.TriggerMode.Async` ("async")

            onRegistered (Callback, optional):
                A function callback that's called on each shard upon function
                registration.
                It is a good place to initialize non-serializable objects such as
                network connections.
                Defaults to None.

        """
        super().__init__(**kwargs)
        self.prefix = prefix
        self.convertToStr = convertToStr
        self.collect = collect
        self.mode = mode
        self.onRegistered = onRegistered

    def add_to(self, function: "PartialGearFunction") -> "ClosedGearFunction":
        """Closes a Gear function with the Register action.

        Args:
            function (PartialGearFunction):
                The partial Gear function to close with the register action.

        Returns:
            ClosedGearFunction:
                A closed Gear event function that is ready to be regisetered on a
                Gears cluster.
        """
        import cloudpickle

        return function.map(lambda x: cloudpickle.dumps(x, protocol=4)).register(
            self.prefix,
            False,
            self.collect,
            mode=self.mode,
            onRegistered=self.onRegistered,
            **self.kwargs,
        )


class Map(Operation):
    """The local Map operation performs the one-to-one (1:1) mapping of records.

    It requires one mapper function.

    Attributes:
        op (redgrease.typing.Mapper):
            The mapper function to map on all input records.
    """

    def __init__(self, op: "optype.Mapper", **kwargs) -> None:
        """Instaniate a Map operation.

        Args:
            op (redgrease.typing.Mapper):
                Function to map on the input records.
                The function must take one argument as input (input record) and
                return something as an output (output record).
        """
        super().__init__(**kwargs)
        self.op = op

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.map(self.op, **self.kwargs)


class FlatMap(Operation):
    """The local FlatMap operation performs one-to-many (1:N) mapping of records.

    It requires one expander function that maps a single input record to potentially
    multiple output records.

    FlatMap is nearly identical to the Map operation in purpose and use.
    Unlike regular mapping, however, when FlatMap returns a sequence / iterator,
    each element in the sequence is turned into a separate output record.

    Attributes:
        op (redgrease.typing.Mapper):
            The mapper function to map on all input records.
    """

    def __init__(self, op: "optype.Expander", **kwargs) -> None:
        """Instaniate a FlatMap operation.

        Args:
            op (redgrease.typing.Expander):
                Function to map on the input records.
                The function must take one argument as input (input record) and
                return an iterable as an output (output records).
        """
        super().__init__(**kwargs)
        self.op = op

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.flatmap(self.op, **self.kwargs)


class ForEach(Operation):
    """The local ForEach operation performs one-to-the-same (1=1) mapping.

    It requires one processor function to perform some work that's related to the
    input record.

    Its output record is a copy of the input, which means anything the callback returns
    is discarded.

    Args:
        op (redgrease.typing.Processor):
            Function to run on the input records.
    """

    def __init__(self, op: "optype.Processor", **kwargs) -> None:
        """Instaniate a ForEach operation.

        Args:
            op (redgrease.typing.Processor):
                Function to run on each of the input records.
                The function must take one argument as input (input record) and
                should not return anything.
        """
        super().__init__(**kwargs)
        self.op = op

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.foreach(self.op, **self.kwargs)


class Filter(Operation):
    """The local Filter operation performs one-to-zero-or-one (1:(0|1)) filtering of
    records.

    It requires a filterer function.

    An input record that yields a falsehood will be discarded and only truthful ones
    will be output.

    Args:
        op (redgrease.typing.Filterer):
            Predicate function to run on the input records.

    """

    def __init__(self, op: "optype.Filterer", **kwargs) -> None:
        """Instaniate a Filter operation.

        Args:
            op (redgrease.typing.Filterer):
                Function to apply on the input records, to decide which ones to keep.
                The function must take one argument as input (input record) and
                return a bool. The input records evaluated to `True` will be kept as
                output records.
        """
        super().__init__(**kwargs)
        self.op = op

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.filter(self.op, **self.kwargs)


class Accumulate(Operation):
    """The local Accumulate operation performs many-to-one mapping (N:1) of records.

    It requires one accumulator function.

    Once input records are exhausted its output is a single record consisting of the
    accumulator's value.

    Args:
        op (redgrease.typing.Accumulator):
            Accumulation function to run on the input records.
    """

    def __init__(self, op: "optype.Accumulator", **kwargs) -> None:
        """Instaniate an Accumulate operation.

        Args:
            op (redgrease.typing.Accumulator):
                Function to to apply on the input records.
                The function must take two arguments as input:
                    - the input record, and
                    - An accumulator value.
                It should aggregate the input record into the accumulator variable,
                which stores the state between the function's invocations.
                The function must return the accumulator's updated value.
        """
        super().__init__(**kwargs)
        self.op = op

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.accumulate(self.op, **self.kwargs)


class LocalGroupBy(Operation):
    """The local LocalGroupBy operation performs many-to-less mapping (N:M) of records.

    The operation requires two functions, an extractor and a reducer.

    The output records consist of the grouping key and its respective reduce value.

    Attributes:
        extractor (redgrease.typing.Extractor):
            Function that extracts the key to group by from input records.

        reducer (redgrease.typing.Reducer):
            Function that reduces the records in each group to an output record.
    """

    def __init__(
        self, extractor: "optype.Extractor", reducer: "optype.Reducer", **kwargs
    ) -> None:
        """Instantiate a LocalGroupBy operator.

        Args:
            extractor (redgrease.typing.Extractor):
                Function to apply on the input records, to extact the grouping key.
                The function must take one argument as input (input record) and
                return a string (key).
                The groups are defined by the value of the key.

            reducer (redgrease.typing.Reducer):
                Function to apply on the records of each group, to reduce to a single
                value (per group).
                The function must take (a) a key, (b) an input record and (c) a
                variable that's called an accumulator.
                It performs similarly to the accumulator callback, with the difference
                being that it maintains an accumulator per reduced key / group.
        """
        super().__init__(**kwargs)
        self.extractor = extractor
        self.reducer = reducer

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.localgroupby(self.extractor, self.reducer, **self.kwargs)


class Limit(Operation):
    """The local Limit operation limits the number of records.

    It accepts two numeric arguments: a starting position in the input records "array"
    and a maximal number of output records.

    Attributes:
        start (int):
            Starting index (0-based) of the input record to start from

        length (int):
            The maximum number of records to let through.
    """

    def __init__(self, length: int, start: int = 0, **kwargs) -> None:
        """Instantiate a Limit operation

        Args:
            length (int):
                The maximum number of records.

            start (int, optional):
                The index of the first input record.
                Defaults to 0.
        """
        super().__init__(**kwargs)
        self.length = length
        self.start = start

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.limit(self.length, self.start, **self.kwargs)


class Collect(Operation):
    """The global Collect operation collects the result records from all of the
    shards to the originating one.
    """

    def __init__(self, **kwargs):
        """Instantiate a Collect operation."""
        super().__init__(**kwargs)

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.collect(**self.kwargs)


class Repartition(Operation):
    """The global Repartition operation repartitions the records by them shuffling
        between shards.

    It accepts a single key extractor function.
    The extracted key is used for computing the record's new placement in the cluster
    (i.e. hash slot).
    The operation then moves the record from its original shard to the new one.

        Attributes:
            extractor (redgreas.typing.Extractor):
                A function deciding the destination shard of an input record.
    """

    def __init__(self, extractor: "optype.Extractor", **kwargs) -> None:
        """Instantiate a Repartition opertation

        Args:
            extractor (redgrease.typing.Extractor):
                Function that takes a record and calculates a key that is used to
                determine the hash slot, and consequently the shard, that the record
                should migrate to to.
                The function must take one argument as input (input record) and
                return a string (key).
                The hash slot, and consequently the destination shard, is determined by
                hthe value of the key.
        """
        super().__init__(**kwargs)
        self.extractor = extractor

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.repartition(self.extractor, **self.kwargs)


class Aggregate(Operation):
    """The Aggregate operation performs many-to-one mapping (N:1) of records.

    Aggregate provides an alternative to the local accumulate operation as it takes
    the partitioning of data into consideration.
    Furthermore, because records are aggregated locally before collection,
    its performance is usually superior.

    It requires a zero value and two accumulator functions for computing the local
    and global aggregates.

    The operation is made of these steps::

    1. The local accumulator is executed locally and initialized with the zero value.
    2. A global collect moves all records to the originating engine.
    3. The global accumulator is executed locally by the originating engine.

    Its output is a single record consisting of the accumulator's global value.

    Attributes:
        zero (Any):
            The initial / zero value for the accumulator variable.

        seqOp (redgrease.typing.Accumulator):
            A local accumulator function, applied locally on each shard.

        combOp (redgrease.typing.Accumulator):
            A global accumulator function, applied on the results of the local
            accumulations.
    """

    def __init__(
        self,
        zero: Any,
        seqOp: "optype.Accumulator",
        combOp: "optype.Accumulator",
        **kwargs,
    ) -> None:
        """Instantiates an Aggregate operation.

        Args:
            zero (Any):
                The initial / zero value of the accumulator variable.

            seqOp (redgrease.typing.Accumulator):
                A function to be applied on each of the input records, locally per
                shard.
                It must take two parameters:
                - an accumulator value, from previous calls
                - an input record
                The functoin aggregates the input into the accumulator variable,
                which stores the state between the function's invocations.
                The function must return the accumulator's updated value.

            combOp (redgrease.typing.Accumulator):
                A function to be applied on each of the aggregated results of the local
                aggregation (i.e. the output of `seqOp`).
                It must take two parameters:
                - an accumulator value, from previous calls
                - an input record
                The functoin aggregates the input into the accumulator variable,
                which stores the state between the function's invocations.
                The function must return the accumulator's updated value.
        """
        super().__init__(**kwargs)
        self.zero = zero
        self.seqOp = seqOp
        self.combOp = combOp

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.aggregate(self.zero, self.seqOp, self.combOp, **self.kwargs)


class AggregateBy(Operation):
    """AggregateBy operation performs many-to-less mapping (N:M) of records.

    It is similar to the Aggregate operation but aggregates per key.
    It requires a an extractor callback, a zero value and two reducers callbacks for
    computing the local and global aggregates.

    The operation is made of these steps::

    1. Extraction of the groups using extractor.
    2. The local reducer is executed locally and initialized with the zero value.
    3. A global repartition operation that uses the extractor.
    4. The global reducer is executed on each shard once it is repartitioned with its
        relevant keys.

    Output list of records, one for each key. The output records consist of the
    grouping key and its respective reducer's value.

    Attributes:
        extractor (redgrease.typing.Extractor):
            Function that extracts the key to group by from input records.

        zero (Any):
            The initial / zero value for the accumulator variable.

        seqOp (redgrease.typing.Accumulator):
            A local accumulator function, applied locally on each shard.

        combOp (redgrease.typing.Accumulator):
            A global accumulator function, applied on the results of the local
            accumulations.
    """

    def __init__(
        self,
        extractor: "optype.Extractor",
        zero: Any,
        seqOp: "optype.Reducer",
        combOp: "optype.Reducer",
        **kwargs,
    ) -> None:
        """Instantiate an AggregateBy operation.

        Args:
            extractor (redgrease.typing.Extractor):
                Function to apply on the input records, to extact the grouping key.
                The function must take one argument as input (input record) and
                return a string (key).
                The groups are defined by the value of the key.

            zero (Any):
                The initial / zero value of the accumulator variable.

            seqOp (redgrease.typing.Accumulator):
                A function to be applied on each of the input records, locally per
                shard and group.
                It must take two parameters:
                - an accumulator value, from previous calls
                - an input record
                The functoin aggregates the input into the accumulator variable,
                which stores the state between the function's invocations.
                The function must return the accumulator's updated value.

            combOp (redgrease.typing.Accumulator):
                A function to be applied on each of the aggregated results of the local
                aggregation (i.e. the output of `seqOp`).
                It must take two parameters:
                - an accumulator value, from previous calls
                - an input record
                The functoin aggregates the input into the accumulator variable,
                which stores the state between the function's invocations.
                The function must return the accumulator's updated value.
        """
        super().__init__(**kwargs)
        self.extractor = extractor
        self.zero = zero
        self.seqOp = seqOp
        self.combOp = combOp

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.aggregateby(
            self.extractor, self.zero, self.seqOp, self.combOp, **self.kwargs
        )


class GroupBy(Operation):
    """GroupBy * operation performs a many-to-less (N:M) grouping of records.
    It is similar to AggregateBy but uses only a global reducer.
    It can be used in cases where locally reducing the data isn't possible.

    The operation requires two functions; an extractor a reducer.

    The operation is made of these steps::

        1. A global repartition operation that uses the extractor.
        2. The reducer is locally invoked.

    Output is a locally-reduced list of records, one for each key.
    The output records consist of the grouping key and its respective accumulator's
    value.

    Attributes:
        extractor (redgrease.typing.Extractor):
            Function that extracts the key to group by from input records.

        reducer (redgrease.typing.Reducer):
            Function that reduces the records of each group to a value
    """

    def __init__(
        self, extractor: "optype.Extractor", reducer: "optype.Reducer", **kwargs
    ) -> None:
        """Instantiate a GroupBy operation.

        Args:
            extractor (redgrease.typing.Extractor):
                Function to apply on the input records, to extact the grouping key.
                The function must take one argument as input (input record) and
                return a string (key).
                The groups are defined by the value of the key.

            reducer (redgrease.typing.Reducer):
                Function to apply on the records of each group, to reduce to a single
                value (per group).
                The function must take (a) a key, (b) an input record and (c) a
                variable that's called an accumulator.
                It performs similarly to the accumulator callback, with the difference
                being that it maintains an accumulator per reduced key / group.
        """
        super().__init__(**kwargs)
        self.extractor = extractor
        self.reducer = reducer

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.groupby(self.extractor, self.reducer, **self.kwargs)


class BatchGroupBy(Operation):
    """BatchGroupBy operation performs a many-to-less (N:M) grouping of records.

    Prefer the GroupBy Operation
    ----------------------------

        Instead of using BatchGroupBy, prefer using the GroupBy operation as it is more
        efficient and performant. Only use BatchGroupBy when the reducer's logic
        requires the full list of records for each input key.

    The operation requires two functions; an extractor a batch reducer.

    The operation is made of these steps::

        1. A global repartition operation that uses the extractor
        2. A local localgroupby operation that uses the batch reducer

    Once finished, the operation locally outputs a record for each key and its
    respective accumulator value.

    Increased memory consumption
    ----------------------------

        Using this operation may cause a substantial increase in memory usage during
        runtime.

    Attributes:
        extractor (redgrease.typing.Extractor):
            Function that extracts the key to group by from input records.

        reducer (redgrease.typing.Reducer):
            Function that reduces the records of each group to a value
    """

    def __init__(
        self, extractor: "optype.Extractor", reducer: "optype.BatchReducer", **kwargs
    ) -> None:
        """Instantiate a BatchGroupBy operation.

        Args:
            extractor (redgrease.typing.Extractor):
                Function to apply on the input records, to extact the grouping key.
                The function must take one argument as input (input record) and
                return a string (key).
                The groups are defined by the value of the key.

            reducer (redgrease.typing.Reducer):
                Function to apply on the records of each group, to reduce to a single
                value (per group).
                The function must take (a) a key, (b) an input record and (c) a
                variable that's called an accumulator.
                It performs similarly to the accumulator callback, with the difference
                being that it maintains an accumulator per reduced key / group.
        """
        super().__init__(**kwargs)
        self.extractor = extractor
        self.reducer = reducer

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.batchgroupby(self.extractor, self.reducer, **self.kwargs)


class Sort(Operation):
    """Sort operation sorts the records.

        It allows to control the sort order.

        The operation is made of the following steps::

            1. A global aggregate operation collects and combines all records.
            2. A local sort is performed on the list.
            3. The list is flatmapped to records.

    Increased memory consumption
    ----------------------------

        Using this operation may cause an increase in memory usage during runtime due
        to the list being copied during the sorting operation.

    Attributes:
        reverse (bool):
            Defines if the sorting orded is descending (`True`) or ascendinng (`False`).
    """

    def __init__(self, reverse: bool = True, **kwargs) -> None:
        """Instantiate a Sort operation.

        Args:
            reverse (bool, optional):
                Sort in descending order (higer to lower).
                Defaults to True.
        """
        super().__init__(**kwargs)
        self.reverse = reverse

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.sort(self.reverse, **self.kwargs)


class Distinct(Operation):
    """The Distinct operation returns distinct records.

    It requires no arguments.

    The operation is made of the following steps::

        1. A aggregate operation locally reduces the records to sets that are then
            collected and unionized globally.
        2. A local flatmap operation turns the set into records.
    """

    def __init__(self, **kwargs):
        """Instantiate a Distinct operation."""
        super().__init__(**kwargs)

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.distinct(**self.kwargs)


class Count(Operation):
    """The Count operation counts the number of input records.

    It requires no arguments.

    The operation is made of an aggregate operation that uses local counting and
    global summing accumulators.
    """

    def __init__(self, **kwargs):
        """Instantiate a Cound operation."""
        super().__init__(**kwargs)

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.count(**self.kwargs)


class CountBy(Operation):
    """The CountBy operation counts the records grouped by key.

    It requires a single extractor function.

    The operation is made of an aggregateby operation that uses local counting and
    global summing accumulators.

    Attributes:
        extractor (redgrease.typing.Extractor):
            Function that extracts the key to group by from input records.
    """

    def __init__(self, extractor: "optype.Extractor", **kwargs) -> None:
        """Instaniate a CountBy operation.

        Args:
            extractor (redgrease.typing.Extractor):
                Function to apply on the input records, to extact the grouping key.
                The function must take one argument as input (input record) and
                return a string (key).
                The groups are defined by the value of the key.
        """
        super().__init__(**kwargs)
        self.extractor = extractor

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.countby(self.extractor, **self.kwargs)


class Avg(Operation):
    """The Avg operation returns the arithmetic average of records.

    It has an optional value extractor function.

    The operation is made of the following steps::

        1. A aggregate operation locally reduces the records to tuples of sum and count
            that are globally combined.
        2. A local map operation calculates the average from the global tupl

    Attributes:
        extractor (redgrease.typing.Extractor):
            Function that extracts the key to group by from input records.
    """

    def __init__(self, extractor: "optype.Extractor", **kwargs) -> None:
        """Instantiate an Avg operation.

        Args:
            extractor (redgrease.typing.Extractor):
                Function to apply on the input records, to extact the grouping key.
                The function must take one argument as input (input record) and
                return a string (key).
                The groups are defined by the value of the key.
        """
        super().__init__(**kwargs)
        self.extractor = extractor

    def add_to(self, function: "PartialGearFunction") -> "PartialGearFunction":
        """Adds the operation to a partial Gear function.

        Args:
            function (PartialGearFunction):
                The partial gear function to add the operation to.

        Returns:
            PartialGearFunction:
                The function with the operation added to the end.
        """
        return function.avg(self.extractor, **self.kwargs)


################################################################################
#                               GearFunctions                                  #
################################################################################


class GearFunction(Generic[T]):
    """Abstract base class for both partial and closed Gear functions.
    The base `GearFunction` class is not intended to be instantiated by API users.

    A GearFunction is a chain of consecutive Operations.

    Attributes:
        operation (Operation):
            The last opertation in the functions chain of operations.

        input_function (PartialGearFunction):
            The function (chain of operations) that provides the input records to the
            `operation`. Two different GearFunctions can share the same `input_function`

        requirements (Iterable[str], optional):
            A set of requirements / dependencies (Python packages) that the operation
            requires in order to execute.
    """

    def __init__(
        self,
        operation: Operation,
        input_function: "PartialGearFunction" = None,
        requirements: Optional[Iterable[str]] = None,
    ) -> None:
        """Instaniate a GearFunction

        Args:
            operation (Operation):
                The last opertation in the functions chain of operations.

            input_function (PartialGearFunction, optional):
                The function (chain of operations) that provides the input records to
                the `operation`.
                Defaults to None.

            requirements (Optional[Iterable[str]], optional):
                A set of requirements / dependencies (Python packages) that the
                operation requires in order to execute.
                Defaults to None.
        """

        self.operation = operation
        self.input_function = input_function
        self.requirements = set(requirements if requirements else [])

        if input_function:
            self.requirements = self.requirements.union(input_function.requirements)

    @property
    def reader(self) -> Optional[str]:
        """The reader type, generating the initial input records to the GearFunction.

        Returns:
            str:
                Either 'KeysReader', 'KeysOnlyReader', 'StreamReader', 'PythonReader',
                    'ShardsIDReader', 'CommandReader' or None (If no reader is defined).
        """
        if isinstance(self.operation, Reader):
            return self.operation.reader

        if self.input_function:
            return self.input_function.reader

        return None

    @property
    def supports_batch_mode(self) -> bool:
        """Indicates if the function can run in Batch-mode, by closing it with a
        `run` action.

        Returns:
            bool:
                `True` if the function supports batch mode, `False` if not.
        """
        return self.reader in [
            sugar.ReaderType.KeysReader,
            sugar.ReaderType.KeysOnlyReader,
            sugar.ReaderType.StreamReader,
            sugar.ReaderType.PythonReader,
            sugar.ReaderType.ShardsIDReader,
        ]

    @property
    def supports_event_mode(self) -> bool:
        """Indicates if the function can run in Event-mode, by closing it with a
        `register` action.

        Returns:
            bool:
                `True` if the function supports event mode, `False` if not.
        """
        return self.reader in [
            sugar.ReaderType.KeysReader,
            sugar.ReaderType.StreamReader,
            sugar.ReaderType.CommandReader,
        ]


class ClosedGearFunction(GearFunction[T]):
    """Closed Gear functions are GearsFunctions that have been 'closed' with a `run`
    action or a `register` action.

    Closed Gear functions cannot add more Operations, but can be executed in a
    Redis Gears cluster.
    """

    def on(
        self,
        gears_server,
        unblocking: bool = False,
        requirements: Iterable[str] = None,
        **kwargs,
    ):
        """Execute the function on a Redis Gear server / cluster.
        This is equivalent to passing the function to `Gears.pyexecute`

        Args:
            gears_server ([type]):
                Redis client / connection object.

            unblocking (bool, optional):
                Execute function unblocking, i.e. asyncronous.
                Defaults to False.

            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

        Returns:
            redgrease.data.ExecutionResult:
                The result of the function, just as `Gears.pyexecute`
        """
        if hasattr(gears_server, "gears"):
            gears_server = gears_server.gears
        if not hasattr(gears_server, "pyexecute"):
            from redgrease.client import Gears

            gears_server = Gears(gears_server)

        return gears_server.pyexecute(
            self, unblocking=unblocking, requirements=requirements, **kwargs
        )


class PartialGearFunction(GearFunction[optype.InputRecord]):
    """A partial Gear function is a Gear function that is not yet 'closed' with a `run`
    action or a `register` action.

    Partial Gear functions can be used to create new partial gear functions by
    applying operations, or it can create a closed Gear function by applying either
    the `run` or `register` action.
    """

    def run(
        self,
        arg: str = None,  # TODO: This can also be a Python generator
        convertToStr: bool = True,
        collect: bool = True,
        # Helpers, all must be None
        # Other Redgrease args
        requirements: Iterable[str] = None,
        on=None,
        # Other Redis Gears args
        **kwargs,
        # TODO: Add all the Reader specific args here
    ) -> ClosedGearFunction["optype.InputRecord"]:
        """Create a closed batch function..

        Batch functions are executed once and exits once the data is
        exhausted by its reader.
        Args:
            arg (str, optional):
                An optional argument that's passed to the reader as its defaultArg.
                It means the following::
                    - A glob-like pattern for the KeysReader and KeysOnlyReader readers.
                    - A key name for the StreamReader reader.
                    - A Python generator for the PythonReader reader.

                Defaults to None.

            convertToStr (bool, optional):
                When `True` adds a map operation to the flow's end that stringifies
                records.
                Defaults to False.

            collect (bool, optional):
                When `True` adds a collect operation to flow's end.
                Defaults to False.

            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

            on (redis.Redis):
                Immedeately execute the function on this Redis Gear server / cluster.

            **kwargs:
                Additional parameters to the run operation.

        Returns:
            Union[ClosedGearFunction, redgrease.data.ExecutionResult]:
                A new closed batch function, if `on` is **not** specified.
                An execution result, if `on` **is** specified.

        Raises:
            TypeError:
                If the function does not support batch mode.
        """
        if not self.supports_batch_mode:
            raise TypeError(f"Batch mode (run) is not supporterd for '{self.reader}'")

        gear_fun: ClosedGearFunction = ClosedGearFunction["optype.InputRecord"](
            Run(arg=arg, convertToStr=convertToStr, collect=collect, **kwargs),
            input_function=self,
            requirements=requirements,
        )

        if redgrease.GEARS_RUNTIME:
            return redgrease.runtime.run(gear_fun, redgrease.GearsBuilder)

        if on:
            return gear_fun.on(on)

        return gear_fun

    def register(  # noqa: C901
        self,
        prefix: str = "*",
        convertToStr: bool = True,
        collect: bool = True,
        # Helpers, all must be None
        mode: str = None,
        onRegistered: "optype.Callback" = None,
        eventTypes: Iterable[str] = None,
        keyTypes: Iterable[str] = None,
        readValue: bool = None,
        batch: int = None,
        duration: int = None,
        onFailedPolcy: str = None,
        onFailedRetryInterval: int = None,
        trimStream: bool = None,
        trigger: str = None,  # Reader Specific: CommandReader
        # Other Redgrease args
        requirements: Iterable[str] = None,
        on=None,
        # Other Redis Gears args
        **kwargs,
        # TODO: Add all the Reader specific args here
    ) -> ClosedGearFunction["optype.InputRecord"]:
        """Create a closed event function.

        Event functions are executed each time an event arrives.
        Each time it is executed, the function operates on the event's
        data and once done is suspended until its future invocations by
        new events.
        Args:
            prefix (str, optional):
                Key prefix pattern to match on.
                Not relevant for 'CommandReader' readers (see 'trigger').
                Defaults to '*'.

            convertToStr (bool, optional):
                When `True` adds a map  operation to the flow's end that stringifies
                records.
                Defaults to True.

            collect (bool, optional):
                When True adds a collect operation to flow's end.
                Defaults to False.

            mode (str, optional):
                The execution mode of the function.
                Can be one of::

                    - 'async': Execution will be asynchronous across the entire
                        cluster.
                    - 'async_local': Execution will be asynchronous and restricted
                        to the handling shard.
                    - 'sync': Execution will be synchronous and local.
                Defaults to 'async'.

            onRegistered (Callback, optional):
                A function that's called on each shard upon function registration.
                It is a good place to initialize non-serializable objects such as
                network connections.
                Defaults to None.

            eventTypes (Iterable[str], optional):
                For KeysReader only.
                A whitelist of event types that trigger execution when the KeysReader
                are used. The list may contain one or more::

                    - Any Redis or module command
                    - Any Redis event

                Defaults to None.

            keyTypes (Iterable[str], optional):
                For KeysReader and KeysOnlyReader only.
                A whitelist of key types that trigger execution when using the
                KeysReader or KeysOnlyReader readers.
                The list may contain one or more from the following::

                    - Redis core types: 'string', 'hash', 'list', 'set', 'zset'
                        or 'stream'
                    - Redis module types: 'module'

                Defaults to None.

            readValue (bool, optional):
                For KeysReader only.
                When `False` the value will not be read, so the 'type' and 'value'
                of the record will be set to None.
                Defaults to True.

            batch (int, optional):
                For StreamReader only.
                The number of new messages that trigger execution.
                Defaults to 1.

            duration  (int, optional):
                For StreamReader only.
                The time to wait before execution is triggered, regardless of the batch
                size (0 for no duration).
                Defaults to 0.

            onFailedPolcy (str, optional):
                For StreamReader only.
                The policy for handling execution failures.
                May be one of::

                    - 'continue': Ignores a failure and continues to the next execution.
                        This is the default policy.
                    - 'abort': Stops further executions.
                    - 'retry': Retries the execution after an interval specified with
                        onFailedRetryInterval (default is one second).

                Defaults to 'continue'.

            onFailedRetryInterval (int, optional):
                For StreamReader only.
                The interval (in milliseconds) in which to retry in case onFailedPolicy
                is 'retry'.
                Defaults to 1.

            trimStream (bool, optional):
                For StreamReader only.
                When True the stream will be trimmed after execution
                Defaults to True.

            trigger (str):
                For 'CommandReader' only, and mandatory.
                The trigger string that will trigger the function.

            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

            on (redis.Redis):
                Immedeately execute the function on this Redis Gear server / cluster.

            **kwargs:
                Additional parameters to the register operation.

        Returns:
            Union[ClosedGearFunction, redgrease.data.ExecutionResult]:
                A new closed event function, if `on` is **not** specified.
                An execution result, if `on` **is** specified.

        Raises:
            TypeError:
                If the function does not support event mode.
        """

        if mode is not None:
            kwargs["mode"] = mode

        if onRegistered is not None:
            kwargs["onRegistered"] = onRegistered

        if not self.supports_event_mode:
            raise TypeError(f"Event mode (run) is not supporterd for '{self.reader}'")

        if eventTypes is not None:
            kwargs["eventTypes"] = list(eventTypes)

        if keyTypes is not None:
            kwargs["keyTypes"] = list(keyTypes)

        if readValue is not None:
            kwargs["readValue"] = readValue

        if batch is not None:
            kwargs["batch"] = batch

        if duration is not None:
            kwargs["duration"] = duration

        if onFailedPolcy is not None:
            kwargs["onFailedPolicy"] = onFailedPolcy

        if onFailedRetryInterval is not None:
            kwargs["onFailedRetryInterval"] = onFailedRetryInterval

        if trimStream is not None:
            kwargs["trimStream"] = trimStream

        if trigger is not None:
            kwargs["trigger"] = trigger

        gear_fun = ClosedGearFunction["optype.InputRecord"](
            Register(
                prefix=prefix,
                convertToStr=convertToStr,
                collect=collect,
                **kwargs,
            ),
            input_function=self,
            requirements=requirements,
        )

        if redgrease.GEARS_RUNTIME:
            return redgrease.runtime.run(gear_fun, redgrease.GearsBuilder)

        if on:
            return gear_fun.on(on)

        return gear_fun

    def map(
        self,
        op: "optype.Mapper[optype.InputRecord, optype.OutputRecord]",
        # Other Redgrease args
        requirements: Iterable[str] = None,
        # Other Redis Gears args
        **kwargs,
    ) -> "PartialGearFunction[optype.OutputRecord]":
        """Instance-local Map operation that performs a one-to-one (1:1) mapping of
        records.

        Args:
            op (redgrease.typing.Mapper):
                Function to map on the input records.
                The function must take one argument as input (input record) and
                return something as an output (output record).

            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

            **kwargs:
                Additional parameters to the Map operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with a Map operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            Map(op=op, **kwargs),
            input_function=self,
            requirements=requirements,
        )

    def flatmap(
        self,
        op: "optype.Expander[optype.InputRecord, optype.OutputRecord]",
        # Other Redgrease args
        requirements: Iterable[str] = None,
        # Other Redis Gears args
        **kwargs,
    ) -> "PartialGearFunction[Iterable[optype.OutputRecord]]":
        """Instance-local FlatMap operation that performs one-to-many (1:N) mapping
        of records.

        Args:
            op (redgrease.typing.Expander):
                Function to map on the input records.
                The function must take one argument as input (input record) and
                return an iterable as an output (output records).

            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

            **kwargs:
                Additional parameters to the FlatMap operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with a FlatMap operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            FlatMap(op=op, **kwargs),
            input_function=self,
            requirements=requirements,
        )

    def foreach(
        self,
        op: "optype.Processor[optype.InputRecord]",
        # Other Redgrease args
        requirements: Iterable[str] = None,
        # Other Redis Gears args
        **kwargs,
    ) -> "PartialGearFunction[optype.InputRecord]":
        """Instance-local ForEach operation performs one-to-the-same (1=1) mapping.

        Args:
            op (redgrease.typing.Processor):
                Function to run on each of the input records.
                The function must take one argument as input (input record) and
                should not return anything.

            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

            **kwargs:
                Additional parameters to the ForEach operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with a ForEach operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            ForEach(op=op, **kwargs),
            input_function=self,
            requirements=requirements,
        )

    def filter(
        self,
        op: "optype.Filterer[optype.InputRecord]",
        # Other Redgrease args
        requirements: Iterable[str] = None,
        # Other Redis Gears args
        **kwargs,
    ) -> "PartialGearFunction[optype.InputRecord]":
        """Instance-local Filter operation performs one-to-zero-or-one (1:bool)
        filtering of records.

        Args:
            op (redgrease.typing.Filterer):
                Function to apply on the input records, to decide which ones to keep.
                The function must take one argument as input (input record) and
                return a bool. The input records evaluated to `True` will be kept as
                output records.

            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

            **kwargs:
                Additional parameters to the Filter operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with a FIlter operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            Filter(op=op, **kwargs),
            input_function=self,
            requirements=requirements,
        )

    def accumulate(
        self,
        op: "optype.Accumulator[T, optype.InputRecord]",
        # Other Redgrease args
        requirements: Iterable[str] = None,
        # Other Redis Gears args
        **kwargs,
    ) -> "PartialGearFunction[T]":
        """Instance-local Accumulate operation performs many-to-one mapping (N:1) of
        records.

        Args:
            op (redgrease.typing.Accumulator):
                Function to to apply on the input records.
                The function must take two arguments as input:
                    - the input record, and
                    - An accumulator value.
                It should aggregate the input record into the accumulator variable,
                which stores the state between the function's invocations.
                The function must return the accumulator's updated value.

            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

            **kwargs:
                Additional parameters to the Accumulate operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with Accumulate operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            Accumulate(op=op, **kwargs),
            input_function=self,
            requirements=requirements,
        )

    def localgroupby(
        self,
        extractor: "optype.Extractor[optype.InputRecord, optype.Key]",
        reducer: "optype.Reducer[optype.Key, T, optype.InputRecord]",
        # Other Redgrease args
        requirements: Iterable[str] = None,
        # Other Redis Gears args
        **kwargs,
    ) -> "PartialGearFunction[Dict[optype.Key, T]]":
        """Instance-local LocalGroupBy operation performs many-to-less mapping (N:M)
        of records.

        Args:
            extractor (redgrease.typing.Extractor):
                Function to apply on the input records, to extact the grouping key.
                The function must take one argument as input (input record) and
                return a string (key).
                The groups are defined by the value of the key.

            reducer (redgrease.typing.Reducer):
                Function to apply on the records of each group, to reduce to a single
                value (per group).
                The function must take (a) a key, (b) an input record and (c) a
                variable that's called an accumulator.
                It performs similarly to the accumulator callback, with the difference
                being that it maintains an accumulator per reduced key / group.

            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

            **kwargs:
                Additional parameters to the LocalGroupBy operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with a LocalGroupBy operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            LocalGroupBy(extractor=extractor, reducer=reducer, **kwargs),
            input_function=self,
            requirements=requirements,
        )

    def limit(
        self,
        length: int,
        start: int = 0,
        # Other Redis Gears args
        **kwargs,
    ) -> "PartialGearFunction[optype.InputRecord]":
        """Instance-local Limit operation limits the number of records.

        Args:
            length (int):
                The maximum number of records.

            start (int, optional):
                The index of the first input record.
                Defaults to 0.

            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

            **kwargs:
                Additional parameters to the Limit operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with a Limit operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            Limit(length=length, start=start, **kwargs),
            input_function=self,
        )

    def collect(self, **kwargs) -> "PartialGearFunction[optype.InputRecord]":
        """Cluster-global Collect operation collects the result records.

        Args:
            **kwargs:
                Additional parameters to the Collect operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with a Collect operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            Collect(**kwargs),
            input_function=self,
        )

    def repartition(
        self,
        extractor: "optype.Extractor[optype.InputRecord, Hashable]",
        # Other Redgrease args
        requirements: Iterable[str] = None,
        # Other Redis Gears args
        **kwargs,
    ) -> "PartialGearFunction[optype.InputRecord]":
        """Cluster-global Repartition operation repartitions the records by shuffling
        them between shards.

        Args:
            extractor (redgrease.typing.Extractor):
                Function that takes a record and calculates a key that is used to
                determine the hash slot, and consequently the shard, that the record
                should migrate to to.
                The function must take one argument as input (input record) and
                return a string (key).
                The hash slot, and consequently the destination shard, is determined by
                hthe value of the key.

            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

            **kwargs:
                Additional parameters to the Repartition operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with a Repartition operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            Repartition(extractor=extractor, **kwargs),
            input_function=self,
            requirements=requirements,
        )

    def aggregate(
        self,
        zero: T,
        seqOp: "optype.Accumulator[T, optype.InputRecord]",
        combOp: "optype.Accumulator[T, T]",
        # Other Redgrease args
        requirements: Iterable[str] = None,
        # Other Redis Gears args
        **kwargs,
    ) -> "PartialGearFunction[T]":
        """Perform aggregation on all the execution data.

        Args:
            zero (Any):
                The initial / zero value of the accumulator variable.

            seqOp (redgrease.typing.Accumulator):
                A function to be applied on each of the input records, locally per
                shard.
                It must take two parameters:
                - an accumulator value, from previous calls
                - an input record
                The functoin aggregates the input into the accumulator variable,
                which stores the state between the function's invocations.
                The function must return the accumulator's updated value.

            combOp (redgrease.typing.Accumulator):
                A function to be applied on each of the aggregated results of the local
                aggregation (i.e. the output of `seqOp`).
                It must take two parameters:
                - an accumulator value, from previous calls
                - an input record
                The functoin aggregates the input into the accumulator variable,
                which stores the state between the function's invocations.
                The function must return the accumulator's updated value.

            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

            **kwargs:
                Additional parameters to the Aggregate operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with a Aggregate operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            Aggregate(zero=zero, seqOp=seqOp, combOp=combOp, **kwargs),
            input_function=self,
            requirements=requirements,
        )

    def aggregateby(
        self,
        extractor: "optype.Extractor[optype.InputRecord, optype.Key]",
        zero: T,
        seqOp: "optype.Reducer[optype.Key, T, optype.InputRecord]",
        combOp: "optype.Reducer[optype.Key, T, T]",
        # Other Redgrease args
        requirements: Iterable[str] = None,
        # Other Redis Gears args
        **kwargs,
    ) -> "PartialGearFunction[Dict[optype.Key, T]]":
        """Like aggregate, but on each key, the key is extracted using the extractor.

        Args:
            extractor (redgrease.typing.Extractor):
                Function to apply on the input records, to extact the grouping key.
                The function must take one argument as input (input record) and
                return a string (key).
                The groups are defined by the value of the key.

            zero (Any):
                The initial / zero value of the accumulator variable.

            seqOp (redgrease.typing.Accumulator):
                A function to be applied on each of the input records, locally per
                shard and group.
                It must take two parameters:
                - an accumulator value, from previous calls
                - an input record
                The functoin aggregates the input into the accumulator variable,
                which stores the state between the function's invocations.
                The function must return the accumulator's updated value.

            combOp (redgrease.typing.Accumulator):
                A function to be applied on each of the aggregated results of the local
                aggregation (i.e. the output of `seqOp`).
                It must take two parameters:
                - an accumulator value, from previous calls
                - an input record
                The functoin aggregates the input into the accumulator variable,
                which stores the state between the function's invocations.
                The function must return the accumulator's updated value.

            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

            **kwargs:
                Additional parameters to the AggregateBy operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with a AggregateBy operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            AggregateBy(
                extractor=extractor, zero=zero, seqOp=seqOp, combOp=combOp, **kwargs
            ),
            input_function=self,
            requirements=requirements,
        )

    def groupby(
        self,
        extractor: "optype.Extractor[optype.InputRecord, optype.Key]",
        reducer: "optype.Reducer[optype.Key, T, optype.InputRecord]",
        # Other Redgrease args
        requirements: Iterable[str] = None,
        # Other Redis Gears args
        **kwargs,
    ) -> "PartialGearFunction[Dict[optype.Key, T]]":
        """Perform a many-to-less (N:M) grouping of records.

        Args:
            extractor (redgrease.typing.Extractor):
                Function to apply on the input records, to extact the grouping key.
                The function must take one argument as input (input record) and
                return a string (key).
                The groups are defined by the value of the key.

            reducer (redgrease.typing.Reducer):
                Function to apply on the records of each group, to reduce to a single
                value (per group).
                The function must take (a) a key, (b) an input record and (c) a
                variable that's called an accumulator.
                It performs similarly to the accumulator callback, with the difference
                being that it maintains an accumulator per reduced key / group.
            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

            **kwargs:
                Additional parameters to the GroupBy operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with a GroupBy operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            GroupBy(extractor=extractor, reducer=reducer, **kwargs),
            input_function=self,
            requirements=requirements,
        )

    def batchgroupby(
        self,
        extractor: "optype.Extractor[optype.InputRecord, optype.Key]",
        reducer: "optype.BatchReducer[optype.Key, T, optype.InputRecord]",
        # Other Redgrease args
        requirements: Iterable[str] = None,
        # Other Redis Gears args
        **kwargs,
    ) -> "PartialGearFunction[Dict[optype.Key, T]]":
        """Many-to-less (N:M) grouping of records.

            Note: Using this operation may cause a substantial increase in memory usage
                during runtime. Consider using the GroupBy

        Args:
            extractor (redgrease.typing.Extractor):
                Function to apply on the input records, to extact the grouping key.
                The function must take one argument as input (input record) and
                return a string (key).
                The groups are defined by the value of the key.

            reducer (redgrease.typing.Reducer):
                Function to apply on the records of each group, to reduce to a single
                value (per group).
                The function must take (a) a key, (b) an input record and (c) a
                variable that's called an accumulator.
                It performs similarly to the accumulator callback, with the difference
                being that it maintains an accumulator per reduced key / group.

            **kwargs:
                Additional parameters to the BatchGroupBy operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with a BatchGroupBy operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            BatchGroupBy(extractor=extractor, reducer=reducer, **kwargs),
            input_function=self,
            requirements=requirements,
        )

    def sort(
        self,
        reverse: bool = True,
        # Other Redgrease args
        requirements: Iterable[str] = None,
        # Other Redis Gears args
        **kwargs,
    ) -> "PartialGearFunction[optype.InputRecord]":
        """Sort the records

        Args:
            reverse (bool, optional):
                Sort in descending order (higer to lower).
                Defaults to True.

            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

            **kwargs:
                Additional parameters to the Sort operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with a Sort operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            Sort(reverse=reverse, **kwargs),
            input_function=self,
            requirements=requirements,
        )

    def distinct(self, **kwargs) -> "PartialGearFunction[optype.InputRecord]":
        """Keep only the distinct values in the data.

        Args:
            **kwargs:
                Additional parameters to the Distinct operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with a Distinct operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            Distinct(**kwargs),
            input_function=self,
        )

    def count(self, **kwargs) -> "PartialGearFunction[int]":
        """Count the number of records in the execution.

        Args:
            **kwargs:
                Additional parameters to the Count operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with a Count operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            Count(**kwargs),
            input_function=self,
        )

    def countby(
        self,
        extractor: "optype.Extractor[optype.InputRecord, Hashable]" = lambda x: str(x),
        # Other Redgrease args
        requirements: Iterable[str] = None,
        # Other Redis Gears args
        **kwargs,
    ) -> "PartialGearFunction[Dict[Hashable, int]]":
        """Counts the records grouped by key.

        Args:
            extractor (redgrease.typing.Extractor):
                Function to apply on the input records, to extact the grouping key.
                The function must take one argument as input (input record) and
                return a string (key).
                The groups are defined by the value of the key.
                Defaults to 'lambda x: str(x)'.

            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

            **kwargs:
                Additional parameters to the CountBy operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with a CountBy operation as last step.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            CountBy(extractor=extractor, **kwargs),
            input_function=self,
            requirements=requirements,
        )

    def avg(
        self,
        extractor: "optype.Extractor[optype.InputRecord, float]" = lambda x: float(
            x if isinstance(x, (int, float, str)) else str(x)
        ),
        # Other Redgrease args
        requirements: Iterable[str] = None,
        # Other Redis Gears args
        **kwargs,
    ) -> "PartialGearFunction[float]":
        """Calculating arithmetic average of the records.

        Args:
            extractor (redgrease.typing.Extractor):
                Function to apply on the input records, to extact the grouping key.
                The function must take one argument as input (input record) and
                return a string (key).
                The groups are defined by the value of the key.
                Defaults to 'lambda x: float(x)'.

            requirements (Iterable[str], optional):
                Additional requirements / dedpendency Python packages.
                Defaults to None.

            **kwargs:
                Additional parameters to the map operation.

        Returns:
            PartialGearFunction:
                A new partial gear function with an avg operation as last step.
                GearsBuilder - The same GearBuilder, but with updated function.

                Note that for GearBuilder this method does **not** return a new
                GearFunction, but instead returns the same GearBuilder, but with its
                internal function updaded.
        """
        return PartialGearFunction(
            Avg(extractor=extractor, **kwargs),
            input_function=self,
            requirements=requirements,
        )
